"""
Copyright 2016-2017 Ellation, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import print_function
import json
import re
import sys
import yaml

import botocore.exceptions

from ef_aws_resolver import EFAwsResolver
from ef_config import EFConfig
from ef_config_resolver import EFConfigResolver
from ef_utils import create_aws_clients, fail, get_account_id, http_get_metadata, whereami
from ef_version_resolver import EFVersionResolver

# pattern to find resolvable symbols - finds innermost nestings
symbol_pattern = re.compile(r'{{\.?([0-9A-Za-z/_,.:\-+=*]+?)}}')
# inverse of SYMBOL_PATTERN, and disallows ':' and ',' from param keys; this is checked in load()
illegal_param_chars = re.compile(r'[^(0-9A-Za-z/_.\-)]')


# Utilities
def get_metadata_or_fail(metadata_key):
  """
  Call get_metadata; halt with fail() if it raises an exception
  """
  try:
    return http_get_metadata(metadata_key)
  except IOError as error:
    fail("Exception in http_get_metadata {} {}".format(metadata_key, repr(error)))


class EFTemplateResolver(object):
  """
  Resolves {{symbol}} style tokens in a file or string from multiple sources:
    - 1. lookups of AWS resource identifiers, such as a security group ID
    - 2. secure credentials
    - 3. version registry for versioned content such as an AMI ID
    - 4. built-in context (such as "ENV")
    - 5. parameters from a parameter file / dictionary of params

  To use:
    from ef_template_resolver import EFTemplateResolver
    # if local (for testing or configuring something other than "self"):
    a = EFTemplateResolver(profile="proto", env="proto3", region="us-east-1", service="myservice")
    # if in a VM, there is no AWS context available:
    a = EFTemplateResolver(env="local", service="myservice")
    # if in EC2:
    a = EFTemplateResolver()
    # then...
    a.load(template, parameters=None, env)
    resolved = a.render() <-- multi-step resolution; returns resolved template
    if (len(a.unresolved_symbols())) > 0:
      print("Template has unresolved symbols")

  1,2,3: AWS values from the live AWS environment, and secure credential storage
    see ef_aws_resolver.py for details

  4: AWS EC2 instance or Lambda context
  These instance-state strings are always available in AWS EC2 and Lambda:
    ACCOUNT - account number in which this EC2 instance or lambda is running
    ACCOUNT_ALIAS - friendly-name alias to the numeric account number
    ENV - prod, staging, proto0..proto3, internal
    ENV_SHORT - ENV with the final digit (if any) removed to make a generic env descriptor
    FUNCTION_NAME - Lambda only
    INSTANCE_ID - EC2 only
    REGION - region in which the instance or lambda is running
    ROLE
    SERVICE - service shortname

  5: Parameter file or a dictionary of parameters
  The strings provided in a parameters file or dictionary can be any other string.
  Lowercase is recommended for template tokens, and matches are case-sensitive.
  If passing a dictionary (or creating one in a file), it's structured like the below.
  Parameter names may not contain the ":" character but can contain: / , - _ .
    {
      "params": {
        "<envA>": {
          "myvarA1": "myvalueA1",
          "myvarA2": "myvalueA2",
          ...
        },
        "<envB>": {
          "myvarA1": "myvalueA1",
          "myvarA2": "myvalueA2",
          ...
  Valid <env> values are:
    "default", "prod", "staging", "proto", "proto<0>".."proto<N>", "localvm", etc.
  <env> sections are evaluated hierarchically, in this order:
    1) "default" - optional; if present is always applied first
    2) the general environment: "prod", "staging", "proto" ("any proto<0>..<N>"), etc.
    3) "proto<N>" - the specific ephemeral environment, if any
  The ephemeral env evaluation happens to allow for example, "proto" to set generic values for all
  proto<N> environments. Then proto<N> values can be applied to specific proto envs if necessary
  to customize for the specific environment.

  """
  # class vars
  __CLIENTS = {}  # boto3 clients
  __AWSR = None   # EFAwsResolver
  __EFCR = None   # EFConfigResolver
  __VR = None     # EFVersionResolver

  def __init__(self,
               profile=None, region=None,  # set both for user access mode
               lambda_context=None,  # set if target is 'self' and this is a lambda
               target_other=False, env=None, service=None,  # set env & service if target_other=True
               skip_symbols={},
               verbose=False
               ):
    """
    Depending on how this is called, access mode (how it logs into AWS) and target (what the
      various context vars report) will vary

    ACCESS MODE - how this logs in to AWS
      user: "running on my laptop or elsewhere with a user credential"
        both 'profile' and 'region' are required
        is always "operating on something else" (TARGET is never "self")
      role: "running in AWS EC2 or Lambda with a role credential"
        do not set profile

    TARGET - what context is reported
      self: "this ec2 instance or lambda is initializing itself"
        assumed for ec2 and lambda, unless target_other=True in the constructor
        never an option for "user" access mode
      other: "this local user, ec2 instance, or lambda is configuring something else"
        always "other" if access mode is "user"
        if access mode is "role", set target_other=True in the constructor
        Constructor must also set 'env' and 'service'

      self:
        lambda_context must be provided if this is a lambda; leave it unset for ec2
        INSTANCE_ID = instance ID if EC2, else None
        FUNCTION_NAME = function name if lambda, else None
        ACCOUNT = numeric account this is running in
        ACCOUNT_ALIAS = named alias of the account this is running in
        ROLE = role this is running as
        ENV = the environment this is running in, from role name
        ENV_SHORT = derived from ENV
        ENV_FULL = fully qualified environment, same as ENV unless env is a global env (mgmt.* or global.*)
        SERVICE = the service this is, from role name
        REGION = region this is running in
      something else:
        INSTANCE_ID = None
        FUNCTION_NAME = None
        ACCOUNT = the numeric account I'm logged into (look up)
        ACCOUNT_ALIAS = the alias of the account i'm logged into (look up)
        ROLE = None
        ENV = the target's environment, passed in from the constructor
        ENV_SHORT = derived from ENV
        ENV_FULL = ENV, with ".<ACCOUNT_ALIAS>" as appropriate
        SERVICE = the service name, passed in from the constructor
        REGION = the region I am in (ec2, lambda) or explicitly set (region= in constructor)

    Collects instance's environment for use in templates:
      {{ACCOUNT}}       - AWS account number
                          CloudFormation can use this or the AWS::AccountID pseudo param
      {{ACCOUNT_ALIAS}} - AWS account alias
      {{ENV}}           - environment: mgmt, prod, staging, proto<N>, etc.
      {{ENV_SHORT}}     - env with <N> or account trimmed: mgmt, prod, staging, proto, etc.
      {{ENV_FULL}}      - env fully qualified: prod, staging, proto<N>, mgmt.<account_alias>, etc.
      {{FUNCTION_NAME}} - only for lambdas
      {{INSTANCE_ID}}   - only for ec2
      {{REGION}}        - the region currently being worked in
                          CloudFormation can use this or the AWS::Region pseudo param
      {{ROLE}}          - the role bound to the ec2 instance or lambda; only for ec2 and lambda
                          CloudFormation: compose role name in template by joining other strings
    """
    # instance vars
    self.verbose = False  # print noisy status if True
    # resolved tokens - only look up symbols once per session. Protect internal names by declaring
    self.resolved = {
        "ACCOUNT": None,
        "ACCOUNT_ALIAS": None,
        "ENV": None,
        "ENV_SHORT": None,
        "ENV_FULL": None,
        "FUNCTION_NAME": None,
        "INSTANCE_ID": None,
        "REGION": None,
        "ROLE": None
    }

    # template and parameters are populated by the load() method as each template is processed
    self.template = None
    # parameters that accompany this template, if any
    self.parameters = {}
    # Sets of symbols found in the current template (only)
    # read back with self.symbols() and self.unresolved_symbols()
    self.symbols = set()
    self.skip_symbols = skip_symbols
    # capture verbosity pref from constructor
    self.verbose = verbose

    # determine ACCESS MODE
    if profile:  # accessing as a user
      target_other = True
      if not region:
        fail("'region' is required with 'profile' for user-mode access")

    where = whereami()

    # require env and service params init() when target is 'other'
    if (target_other or where == "virtualbox-kvm") and (env is None or service is None):
      fail("'env' and 'service' must be set when target is 'other' or running in " + where)

    if target_other or profile:
      self.resolved["REGION"] = region
    # lambda initializing self
    elif lambda_context:
      self.resolved["REGION"] = lambda_context.invoked_function_arn.split(":")[3]
    # ec2 initializing self
    else:
      self.resolved["REGION"] = get_metadata_or_fail("placement/availability-zone/")[:-1]

    # Create clients - if accessing by role, profile should be None
    clients = [
      "cloudformation",
      "cloudfront",
      "cognito-identity",
      "cognito-idp",
      "dynamodb",
      "ec2",
      "ecr",
      "elbv2",
      "iam",
      "kms",
      "lambda",
      "ram",
      "route53",
      "s3",
      "sts",
      "waf"
    ]
    try:
      EFTemplateResolver.__CLIENTS = create_aws_clients(self.resolved["REGION"], profile, *clients)
    except RuntimeError as error:
      fail("Exception logging in with Session()", error)

    # Create EFAwsResolver object for interactive lookups
    EFTemplateResolver.__AWSR = EFAwsResolver(EFTemplateResolver.__CLIENTS)
    # Create EFConfigResolver object for ef tooling config lookups
    EFTemplateResolver.__EFCR = EFConfigResolver()
    # Create EFVersionResolver object for version lookups
    EFTemplateResolver.__VR = EFVersionResolver(EFTemplateResolver.__CLIENTS)

    # Set the internal parameter values for aws
    # self-configuring lambda
    if (not target_other) and lambda_context:
      arn_split = lambda_context.invoked_function_arn.split(":")
      self.resolved["ACCOUNT"] = arn_split[4]
      self.resolved["FUNCTION_NAME"] = arn_split[6]
      try:
        lambda_desc = EFTemplateResolver.__CLIENTS["lambda"].get_function()
      except:
        fail("Exception in get_function: ", sys.exc_info())
      self.resolved["ROLE"] = lambda_desc["Configuration"]["Role"]
      env = re.search("^({})-".format(EFConfig.VALID_ENV_REGEX), self.resolved["ROLE"])
      if not env:
        fail("Did not find environment in lambda function name.")
      self.resolved["ENV"] = env.group(1)
      parsed_service = re.search(self.resolved["ENV"] + "-(.*?)-lambda", self.resolved["ROLE"])
      if parsed_service:
        self.resolved["SERVICE"] = parsed_service.group(1)

    # self-configuring EC2
    elif (not target_other) and (not lambda_context):
      self.resolved["INSTANCE_ID"] = get_metadata_or_fail('instance-id')
      profile_arn = str(json.loads(http_get_metadata('iam/info'))["InstanceProfileArn"])
      self.resolved["ACCOUNT"] = profile_arn.split(":")[4]
      self.resolved["ROLE"] = profile_arn.split("/")[-1]
      self.resolved["ENV"] = self.resolved["ROLE"].split("-")[0]
      self.resolved["SERVICE"] = "-".join(self.resolved["ROLE"].split("-")[1:])

    # target is "other"
    else:
      try:
        if whereami() == "ec2":
          self.resolved["ACCOUNT"] = str(json.loads(http_get_metadata('iam/info'))["InstanceProfileArn"].split(":")[4])
        else:
          self.resolved["ACCOUNT"] = get_account_id(EFTemplateResolver.__CLIENTS["sts"])
      except botocore.exceptions.ClientError as error:
        fail("Exception in get_user()", error)
      self.resolved["ENV"] = env
      self.resolved["SERVICE"] = service

    # ACCOUNT_ALIAS is resolved consistently for access modes and targets other than virtualbox
    try:
      self.resolved["ACCOUNT_ALIAS"] = EFTemplateResolver.__CLIENTS["iam"].list_account_aliases()["AccountAliases"][0]
    except botocore.exceptions.ClientError as error:
      fail("Exception in list_account_aliases", error)

    # ENV_SHORT is resolved the same way for all access modes and targets
    self.resolved["ENV_SHORT"] = self.resolved["ENV"].strip(".0123456789")

    # ENV_FULL is resolved the same way for all access modes and targets, depending on previously-resolved values
    if self.resolved["ENV"] in EFConfig.ACCOUNT_SCOPED_ENVS:
      self.resolved["ENV_FULL"] = "{}.{}".format(self.resolved["ENV"], self.resolved["ACCOUNT_ALIAS"])
    else:
      self.resolved["ENV_FULL"] = self.resolved["ENV"]

    if self.verbose:
      print(repr(self.resolved), file=sys.stderr)

  def load(self, template, parameters=None):
    """
    'template'
    Loads template text from a 'string' or 'file' type
    Template text contains {{TOKEN}} symbols to be replaced

    'parameters'
    parameters contains environment-specific sections as discussed in the class documentation.
    the 'parameters' arg can be None, a 'string', 'file', or 'dictionary'

    Whether from a string or file, or already in a dictionary, parameters must follow the
    logical format documented in the class docstring.
    if 'parameters' is omitted, template resolution will proceed with AWS, credential, and
    version lookups.
    """
    # load template
    if isinstance(template, str):
      self.template = template
    elif isinstance(template, file):
      try:
        self.template = template.read()
        template.close()
      except IOError as error:
        fail("Exception loading template from file: ", error)
    else:
      fail("Unknown type loading template; expected string or file: " + type(template))

    # load parameters, if any
    if parameters:
      if isinstance(parameters, str):
        try:
          self.parameters = yaml.safe_load(parameters)
        except ValueError as error:
          fail("Exception loading parameters from string: ", error)
      elif isinstance(parameters, file):
        try:
          self.parameters = yaml.safe_load(parameters)
          parameters.close()
        except ValueError as error:
          fail("Exception loading parameters from file: {}".format(error), sys.exc_info())
      elif isinstance(parameters, dict):
        self.parameters = parameters
      else:
        fail("Unknown type loading parameters; expected string, file, or dict: " + type(parameters))
      # sanity check the loaded parameters
      if "params" not in self.parameters:
        fail("'params' field not found in parameters")
      # just the params, please
      self.parameters = self.parameters["params"]
      # are all the keys valid (must have legal characters)
      for k in set().union(*(self.parameters[d].keys() for d in self.parameters.keys())):
        invalid_char = illegal_param_chars.search(k)
        if invalid_char:
          fail("illegal character: '" + invalid_char.group(0) + "' in parameter key: " + k)

  def search_parameters(self, symbol):
    """
    Hierarchically searches for 'symbol' in the parameters blob if there is one (would have
    been retrieved by 'load()'). Order is: default, <env_short>, <env>
    Returns
      Hierarchically resolved value for 'symbol', or None if a match is not found or there are no parameters
    """
    if not self.parameters:
      return None
    # Hierarchically lookup the key
    result = None
    if "default" in self.parameters and symbol in self.parameters["default"]:
      result = self.parameters["default"][symbol]
    if self.resolved["ENV_SHORT"] in self.parameters and symbol in self.parameters[self.resolved["ENV_SHORT"]]:
      result = self.parameters[self.resolved["ENV_SHORT"]][symbol]
    # This lookup is redundant when env_short == env_full, but it's also cheap. Also handles the case for mgmt.<account_alias>
    if self.resolved["ENV_FULL"] in self.parameters and symbol in self.parameters[self.resolved["ENV_FULL"]]:
      result = self.parameters[self.resolved["ENV_FULL"]][symbol]
    return result

  def render(self):
    """
    Find {{}} tokens; resolve then replace them as described elsewhere
    Resolution is multi-pass: tokens may be nested to form parts of other tokens.

    Token search steps when resolving symbols
      - 1. lookups of AWS resource identifiers, such as a security group ID
      - 2. secure credentials
      - 3. version registry for versioned content such as an AMI ID
      - 4. built-in context (such as "ENV")
      - 5. parameters from a parameter file / dictionary of params
    """
    # Ensure that our symbols are clean and are not from a previous template that was rendered
    self.symbols = set()
    # Until all symbols are resolved or it is determined that some cannot be resolved, repeat:
    go_again = True
    while go_again:
      go_again = False  # if at least one symbol isn't resolved in a pass, stop
      # Gather all resolvable symbols in the template
      template_symbols = set(symbol_pattern.findall(self.template))
      self.symbols.update(template_symbols)  # include this pass's symbols in full set
      # resolve and replace symbols
      for symbol in template_symbols:
        resolved_symbol = None

        # Don't resolve symbols that are provided as skippable
        if symbol.split(',')[0] in self.skip_symbols:
          resolved_symbol = "SKIPPED_SYMBOL"
        # Lookups in AWS, only if we have an EFAwsResolver
        elif symbol[:4] == "aws:" and EFTemplateResolver.__AWSR:
          resolved_symbol = EFTemplateResolver.__AWSR.lookup(symbol[4:])
        # Lookups in credentials
        elif symbol[:12] == "credentials:":
          pass  #TODO
        elif symbol[:9] == "efconfig:":
          resolved_symbol = EFTemplateResolver.__EFCR.lookup(symbol[9:])
        elif symbol[:8] == "version:":
          resolved_symbol = EFTemplateResolver.__VR.lookup(symbol[8:])
          if not resolved_symbol:
            print("WARNING: Lookup failed for {{%s}} - placeholder value of 'NONE' used in rendered template" % symbol)
            resolved_symbol = "NONE"
        else:
          # 1. context - these are already in the resolved table
          # self.resolved[symbol] may have value=None; use has_key tell "resolved w/value=None" from "not resolved"
          # these may be "global" symbols such like "ENV", "ACCOUNT", etc.
          if symbol in self.resolved:
            resolved_symbol = self.resolved[symbol]
          # 2. parameters
          if not resolved_symbol:
            resolved_symbol = self.search_parameters(symbol)
        # if symbol was resolved, replace it everywhere
        if resolved_symbol is not None:
          if isinstance(resolved_symbol, list):
            # Using old style of string formatting here due to str.format() interaction with curly braces
            self.template = re.sub(r'{{\.?%s}}' % re.escape(symbol), "\n".join(resolved_symbol), self.template)
          else:
            self.template = re.sub(r'{{\.?%s}}' % re.escape(symbol), resolved_symbol, self.template)
          go_again = True
    return self.template

  def unresolved_symbols(self):
    return set(symbol_pattern.findall(self.template))

  def count_braces(self):
    """
    returns a count of "{{" and "}}" in the template, as (N_left_braces, N_right_braces)
    Useful to check after resolve() has run, to infer that template has an error since no {{ or }}
    should be present in the template after resolve()
    """
    n_left = len(re.findall("{{", self.template))
    n_right = len(re.findall("}}", self.template))
    return n_left, n_right

  def resolved_ok(self):
    """
    Shortcut to testing unresolved_symbols and count_braces separately.
    Returns false if there are unresolved symbols or {{ or }} braces remaining, true otherwise
    """
    left_braces, right_braces = self.count_braces()
    return len(self.unresolved_symbols()) == left_braces == right_braces == 0



from copy import deepcopy
from itertools import chain
import logging

from ef_config import EFConfig
from ef_conf_utils import get_account_alias
from ef_utils import kms_decrypt
from newrelic_interface import NewRelic, AlertPolicy


logging.basicConfig(level=logging.INFO, Force=True)
logger = logging.getLogger(__name__)


class NewRelicAlerts(object):

  def __init__(self, context, clients):
    self.config = EFConfig.PLUGINS['newrelic']
    self.ec2_conditions = self.config.get('ec2_alert_conditions', {})
    self.ecs_conditions = self.config.get('ecs_alert_conditions', {})
    self.local_alert_nrql_conditions = self.config.get('alert_nrql_conditions', {})
    self.lambda_alert_conditions = self.config.get('lambda_alert_conditions', {})
    self.local_alert_apm_conditions = self.config.get('apm_metric_alert_conditions', {})
    self.all_notification_channels = self.config.get('env_notification_map', {})
    self.opsgenie_api_key = self.config["opsgenie_api_key"]
    self.context, self.clients = context, clients

  @classmethod
  def replace_symbols(cls, condition_obj, symbols):
    def convert_string(string):
      for symbol in symbols:
        if symbol in string:
          string = string.replace("{{" + symbol + "}}", str(symbols[symbol]))
          if isinstance(symbols[symbol], int):
            string = int(string)
      return string

    if isinstance(condition_obj, dict):
      for inner_key, inner_val in condition_obj.items():
        condition_obj[inner_key] = cls.replace_symbols(inner_val, symbols)
    elif isinstance(condition_obj, list):
      for i in range(len(condition_obj)):
        condition_obj[i] = cls.replace_symbols(condition_obj[i], symbols)
    elif isinstance(condition_obj, str):
      condition_obj = convert_string(condition_obj)
    return condition_obj

  @property
  def admin_token(self):
    admin_token = self.config.get('admin_token', "")
    admin_token_map = self.config.get('admin_token_map', "")

    if admin_token:
      plain_token = admin_token
    else:
      aws_account_alias = get_account_alias(self.context.env)
      plain_token = admin_token_map.get(aws_account_alias)

    if plain_token is None:
      raise KeyError("NewRelic Admin Token not defined for {}".format(aws_account_alias))

    if self.config['token_kms_encrypted']:
      return kms_decrypt(self.clients['kms'], plain_token).plaintext

    return plain_token

  def create_alert_policy(self, policy):
    if not self.newrelic.alert_policy_exists(policy.name):
      self.newrelic.create_alert_policy(policy.name)
      logger.info("create alert policy {}".format(policy.name))
    return policy

  def populate_alert_policy_values(self, policy, service_type):
    policy.id = next(alert['id'] for alert in self.newrelic.all_alert_policies if alert['name'] == policy.name)
    policy.notification_channels = self.all_notification_channels[self.context.env]
    policy.remote_conditions = self.newrelic.get_policy_alert_conditions(policy.id)
    if service_type in ['aws_ec2', 'http_service']:
      policy.config_conditions = deepcopy(self.ec2_conditions)
    elif service_type in ['aws_ecs', 'aws_ecs_http']:
      policy.config_conditions = deepcopy(self.ecs_conditions)
    elif service_type in ['aws_lambda']:
      policy.config_conditions = deepcopy(self.lambda_alert_conditions)
    policy.remote_alert_nrql_conditions = self.newrelic.get_policy_alert_nrql_conditions(policy.id)
    policy.local_alert_nrql_conditions = deepcopy(self.local_alert_nrql_conditions)
    policy.remote_alert_apm_conditions = self.newrelic.get_policy_alert_apm_conditions(policy.id)
    policy.local_alert_apm_conditions = deepcopy(self.local_alert_apm_conditions)

  def delete_conditions_not_matching_config_values(self, policy):
    # Remove conditions with values that differ from config
    for condition in policy.remote_conditions:
      if condition['name'] in policy.config_conditions:
        config_condition = policy.config_conditions[condition['name']]
        for k, v in config_condition.items():
          if k in condition and condition[k] != v:
            self.newrelic.delete_policy_alert_condition(condition['id'])
            policy.remote_conditions = self.newrelic.get_policy_alert_conditions(policy.id)
            logger.info("delete condition {} with value {} from policy {}. ".format(condition['name'], condition[k], policy.name) +
                        "due to different current local config value {}".format(v))
            break

    return policy

  def create_infra_alert_conditions(self, policy):
    # Create alert conditions for policies
    remote_conditions = set([condition['name'] for condition in policy.remote_conditions])
    conditions_to_create = set(policy.config_conditions.keys()).difference(remote_conditions)

    for key in conditions_to_create:
      logger.info("create condition {} for policy {}".format(key, policy.name))
      self.newrelic.create_alert_condition(policy.config_conditions[key])

  def update_cloudfront_policy(self):
    # Update Cloudfront alert policies
    logger.info("update cloudfront alert policy")
    cloudfront = self.clients['cloudfront']
    pages = cloudfront.get_paginator('list_distributions').paginate()
    queue = []
    map_function = lambda x: (x['Id'], ', '.join(x['Aliases']['Items']))
    filtered_4xx_distributions = []
    for page in pages:
      for distribution in page['DistributionList']['Items']:
        tag_response = cloudfront.list_tags_for_resource(Resource=distribution['ARN'])
        # Remove from distribution list those that do not have tag nr_monitoring set to enabled
        for tag in tag_response['Tags']['Items']:
          if tag['Key'].lower() == "newrelic_4xx_monitoring" and tag['Value'].lower() == "enabled":
            filtered_4xx_distributions.append(distribution['Id'])
            break
      queue.extend(map(map_function, page['DistributionList']['Items']))

    # Create policy conditions
    meta = lambda error_rate, value, distribution_id, name, policy_id: {
      'select_value': 'provider.{}.Average'.format(error_rate),
      'comparison': 'above',
      'critical_threshold': {
        'duration_minutes': 5,
        'time_function': 'all',
        'value': value
      },
      'enabled': True,
      'event_type': 'LoadBalancerSample',
      'filter': {
        'and': [{
          'is': {
            'provider.distributionId': distribution_id
          }
        }]
      },
      'name': name,
      'integration_provider': 'CloudFrontDistribution',
      'policy_id': policy_id,
      'type': 'infra_metric',
    }

    policy = AlertPolicy(env=self.context.env, service='cloudfront')

    # Create cloudfront alert policy if it doesn't already exist
    if not self.newrelic.alert_policy_exists(policy.name):
      self.newrelic.create_alert_policy(policy.name)
      logger.info("create alert policy {}".format(policy.name))

    self.populate_alert_policy_values(policy, 'aws_fixture')
    self.add_alert_policy_to_notification_channels(policy)

    conditions = {}
    for id, alias in queue:
      if id in filtered_4xx_distributions:
        conditions['4xx Average {}'.format(alias)] = meta(
          'error4xxErrorRate', 10, id, '4xx Average {}'.format(alias), policy.id)
      conditions['5xx Average {}'.format(alias)] = meta(
        'error5xxErrorRate', 5, id, '5xx Average {}'.format(alias), policy.id)

    policy.config_conditions = deepcopy(conditions)
    # Infra alert conditions
    policy = self.delete_conditions_not_matching_config_values(policy)
    self.create_infra_alert_conditions(policy)

  def add_alert_policy_to_notification_channels(self, policy):
    # Add alert policy to notification channels if missing
    for channel in self.newrelic.all_notification_channels:
      if channel['name'] in policy.notification_channels and policy.id not in channel['links']['policy_ids']:
        self.newrelic.add_policy_channels(policy.id, [channel['id']])
        logger.info("add channel_ids {} to policy {}".format(policy.name, channel['id']))

  def add_policy_to_opsgenie_channel(self, policy, team_name):
    team_channel = self.newrelic.get_notification_channel_by_name(team_name)
    if not team_channel or team_channel['type'] != 'opsgenie':
      team_channel = self.newrelic.create_opsgenie_alert_channel(
        name=team_name,
        api_key=self.opsgenie_api_key,
        teams=[team_name]
      )

    if policy.id in team_channel['links']['policy_ids']:
      return

    chan_id = team_channel['id']

    self.newrelic.add_policy_channels(policy.id, [chan_id])
    logger.info("add OpsGenie channel id:%s for team %s to policy %s", chan_id, team_name, policy.name)

  def replace_symbols_in_condition(self, policy):
    # Replace symbols in config alert conditions
    for key, value in policy.config_conditions.items():
      policy.config_conditions[key] = self.replace_symbols(value, policy.symbols)

    # Replace symbols in config alert conditions
    for key, value in policy.local_alert_nrql_conditions.items():
      policy.local_alert_nrql_conditions[key] = self.replace_symbols(value, policy.symbols)

  def override_infra_alert_condition_values(self, policy, service_alert_overrides):
    # Update policy.config_conditions with overrides from service_registry
    for condition_name, override_obj in service_alert_overrides.items():
      if condition_name in policy.config_conditions.keys():
        for override_key, override_value in override_obj.items():
          if isinstance(override_value, dict):
            for inner_key, inner_val in override_value.items():
              policy.config_conditions[condition_name][override_key][inner_key] = inner_val
          else:
            policy.config_conditions[condition_name][override_key] = override_value
    logger.debug("Policy {} alert condition values:\n{}".format(policy.name, policy.config_conditions))

    return policy

  def update_alert_nrql_condition_if_different(self, local_alert_nrql_condition, policy):
    for remote_alert_nrql_condition in policy.remote_alert_nrql_conditions:
      if remote_alert_nrql_condition["name"] == local_alert_nrql_condition["name"]:
        relevant_fields = ["nrql", "terms", "enabled", "value_function", "violation_time_limit_seconds"]
        for key in relevant_fields:
          if local_alert_nrql_condition[key] != remote_alert_nrql_condition[key]:
            logger.info("Local alert nrql condition differs from remote alert nrql condition for {}-{}. Updating remote.".format(policy.env, policy.service))
            logger.info("Local: {}\nRemote: {}".format(local_alert_nrql_condition, remote_alert_nrql_condition))
            self.newrelic.put_policy_alert_nrql_condition(remote_alert_nrql_condition["id"], local_alert_nrql_condition)
            break

  def update_alert_apm_condition_if_different(self, local_alert_apm_condition, policy):
    for remote_alert_apm_condition in policy.remote_alert_apm_conditions:
      if remote_alert_apm_condition["name"] == local_alert_apm_condition["name"]:
        # Add fields that exist in the remote alert condition object but not in the local alert condition object.
        # This is done so that we can test equality.
        local_alert_apm_condition['id'] = remote_alert_apm_condition['id']
        local_alert_apm_condition['type'] = remote_alert_apm_condition['type']
        local_alert_apm_condition['entities'] = remote_alert_apm_condition['entities']

        local_alert_apm_condition['terms'].sort(key=lambda x: x['priority'])
        remote_alert_apm_condition['terms'].sort(key=lambda x: x['priority'])

        if local_alert_apm_condition != remote_alert_apm_condition:
          logger.info("Local alert apm condition differs from remote alert apm condition for {}-{}. Updating remote.".format(policy.env, policy.service))
          logger.debug("Local: {}\nRemote: {}".format(local_alert_apm_condition, remote_alert_apm_condition))
          self.newrelic.put_policy_alert_apm_condition(remote_alert_apm_condition["id"], local_alert_apm_condition)

  def override_alert_apm_condition_values(self, policy, service_alert_overrides):
    # Update policy.config_conditions with overrides from service_registry
    for condition_name, override_obj in service_alert_overrides.items():
      if condition_name in policy.local_alert_apm_conditions.keys():
        for override_key, override_value in override_obj.items():
          if isinstance(override_value, dict):
            for inner_key, inner_val in override_value.items():
              policy.local_alert_apm_conditions[condition_name][override_key][inner_key] = inner_val
          else:
            policy.local_alert_apm_conditions[condition_name][override_key] = override_value
    logger.debug("Policy {} APM alert condition values:\n{}".format(policy.name, policy.local_alert_apm_conditions))

    return policy

  def update_application_services_policies(self):
    for service_name, service_config in self.context.service_registry.iter_services(service_group="application_services"):
      service_environments = service_config['environments']
      service_alert_overrides = service_config.get('alerts', {})
      opsgenie_enabled = service_alert_overrides.get('opsgenie_enabled', True)
      opsgenie_team = service_config.get("team_opsgenie", "")
      service_type = service_config['type']

      if service_type not in ['aws_ec2', 'aws_ecs', 'aws_ecs_http', 'http_service']:
        continue

      logger.info("Service: {}".format(service_name))

      if self.context.env in service_environments:
        policy = AlertPolicy(env=self.context.env, service=service_name)

        logger.debug("Policy ID is {}".format(policy.id))

        # Create service alert policy if it doesn't already exist
        if not self.newrelic.alert_policy_exists(policy.name):
          self.newrelic.create_alert_policy(policy.name)
          logger.info("create alert policy {}".format(policy.name))

        # Update AlertPolicy object
        self.populate_alert_policy_values(policy, service_type)
        self.add_alert_policy_to_notification_channels(policy)
        self.replace_symbols_in_condition(policy)

        # Configure Opsgenie notifications for services running in the production account
        try:
          prod_account = EFConfig.ENV_ACCOUNT_MAP['prod']
          if (self.context.env in ["prod", "global.{}".format(prod_account), "mgmt.{}".format(prod_account)]
              and opsgenie_enabled and opsgenie_team):
            self.add_policy_to_opsgenie_channel(policy, opsgenie_team)
          elif not opsgenie_enabled or not opsgenie_team:
            logger.warning("Not adding opsgenie_channel {} for service {} alert policy.".format(opsgenie_team, service_name))
        except KeyError:
          pass

        # Infra alert conditions
        policy = self.override_infra_alert_condition_values(policy, service_alert_overrides)
        policy = self.delete_conditions_not_matching_config_values(policy)
        self.create_infra_alert_conditions(policy)

        # NRQL alert conditions
        remote_alert_nrql_condition_names = [remote_alert_nrql_condition['name'] for remote_alert_nrql_condition in policy.remote_alert_nrql_conditions]
        for condition_name, condition_value in policy.local_alert_nrql_conditions.items():
          if unicode(condition_name, "utf-8") not in remote_alert_nrql_condition_names:
            logger.info("Creating alert nrql condition {} for service {}".format(condition_name, service_name))
            self.newrelic.create_alert_nrql_condition(policy.id, condition_value)
          else:
            self.update_alert_nrql_condition_if_different(condition_value, policy)

        # APM alert conditions
        policy = self.override_alert_apm_condition_values(policy, service_alert_overrides)
        remote_alert_apm_condition_names = [remote_alert_apm_condition['name'] for remote_alert_apm_condition in policy.remote_alert_apm_conditions]
        for condition_name, condition_value in policy.local_alert_apm_conditions.items():
          if unicode(condition_name, "utf-8") not in remote_alert_apm_condition_names:
            applications = self.newrelic.get_applications(application_name=policy.name)

            if not len(applications):
              logger.info('No applications hosted for this policy: {}. Skip creating any APM alert'.format(policy.name))
              continue

            condition_value['entities'] = [applications[0]['id']]
            logger.info("Creating alert apm condition {} for service {}".format(condition_name, service_name))
            self.newrelic.create_alert_apm_condition(policy.id, condition_value)
          else:
            self.update_alert_apm_condition_if_different(condition_value, policy)

  def update_lambda_policies(self):
    platform_services = self.context.service_registry.iter_services(service_group="platform_services")
    application_services = self.context.service_registry.iter_services(service_group="application_services")
    for service_name, service_config in chain(platform_services, application_services):
      service_environments = service_config['environments']
      service_alert_overrides = service_config.get('alerts', {})
      opsgenie_enabled = service_alert_overrides.get('opsgenie_enabled', True)
      opsgenie_team = service_config.get("team_opsgenie", "")
      service_type = service_config['type']

      if service_type != 'aws_lambda':
        continue

      logger.info("Service: {}".format(service_name))

      if self.context.env in service_environments:
        policy = AlertPolicy(env=self.context.env, service=service_name)

        logger.debug("Policy ID is {}".format(policy.id))

        # Create service alert policy if it doesn't already exist
        if not self.newrelic.alert_policy_exists(policy.name):
          self.newrelic.create_alert_policy(policy.name)
          logger.info("create alert policy {}".format(policy.name))

        # Update AlertPolicy object
        self.populate_alert_policy_values(policy, service_type)
        self.add_alert_policy_to_notification_channels(policy)
        self.replace_symbols_in_condition(policy)

        # Configure Opsgenie notifications for services running in the production account
        try:
          prod_account = EFConfig.ENV_ACCOUNT_MAP['prod']
          if (self.context.env in ["prod", "global.{}".format(prod_account), "mgmt.{}".format(prod_account)]
              and opsgenie_enabled and opsgenie_team):
            self.add_policy_to_opsgenie_channel(policy, opsgenie_team)
          elif not opsgenie_enabled or not opsgenie_team:
            logger.warning("Not adding opsgenie_channel {} for service {} alert policy.".format(opsgenie_team, service_name))
        except KeyError:
          pass

        policy = self.override_infra_alert_condition_values(policy, service_alert_overrides)
        policy = self.delete_conditions_not_matching_config_values(policy)
        self.create_infra_alert_conditions(policy)

  def run(self):
    if self.context.env in self.all_notification_channels.keys():
      self.newrelic = NewRelic(self.admin_token)
      self.update_application_services_policies()
      self.update_lambda_policies()

      if self.context.env in ["prod"]:
        self.update_cloudfront_policy()



"""
Copyright 2016-2017 Ellation, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import print_function

from botocore.client import ClientError

from ef_config import EFConfig

class EFVersionResolver(object):
  """
  Resolves "the correct version" of various things, including AMIs.
  In fact, all this resolves right now is /the latest version/ of an AMI.

  Syntax:
    <thing>,<env>/<service_name>
  Example:
    ami-id,prod/ess

  <thing> can be:
    ami-id: the ID of the AMI for <service_name>; ami must be named "<service_name>-release"
    (others will come in the future)

  In a template:
    {{version:<thing>,<env>/<service>}}
    {{version:ami-id,staging/ess}} <-- gets the ID of the ami for the ess service in the current environment

  Requires these clients:
    clients["ec2", "iam", "lambda", "s3"]
  """

  # dictionary of boto3 clients: {"ec2":ec2_client, ...}
  __CLIENTS = {}

  def _s3_get(self, env, service, key):
    s3_key = "{}/{}/{}".format(service, env, key)
    try:
      s3_object = EFVersionResolver.__CLIENTS["s3"].get_object(
        Bucket = EFConfig.S3_VERSION_BUCKET,
        Key = s3_key
      )
    except ClientError as e:
      # If object was legit not found, return None. This is not a failure.
      # Allows precheck to continue for bootstrapping, and the template resolver expects None in this case
      response_code = e.response["ResponseMetadata"]["HTTPStatusCode"]
      if response_code == 404:
        return None
      # Otherwise, an unexpected issue occurred. Stop with error
      raise
    # Else get the value and decide what to do with it
    return s3_object["Body"].read().decode(EFConfig.S3_VERSION_CONTENT_ENCODING)

  def lookup(self, token):
    """
    Return key version if found, None otherwise
    Lookup should look like this:
      pattern:
      <key>,<env>/<service>
      example:
      ami-id,staging/core
    """
    # get search key and env/service from token
    try:
      key, envservice = token.split(",")
    except ValueError:
      return None
    # get env, service from value
    try:
      env, service = envservice.split("/")
    except ValueError as e:
      raise RuntimeError("Request:{} can't resolve to env, service. {}".format(envservice, e))

    return self._s3_get(env, service, key)

  def __init__(self, clients):
    """
    ARGS
      clients - dictionary of ready-to-go boto3 clients using aws prefixes:
      expected: clients["ec2", "iam", "lambda", "s3"]
    """
    EFVersionResolver.__CLIENTS = clients



#!/usr/bin/env python
from __future__ import print_function

import argparse
import json
import math
import os
import re
import subprocess
import sys
import time

import botocore.exceptions

from ef_config import EFConfig
from ef_context import EFContext
from ef_service_registry import EFServiceRegistry
from ef_template_resolver import EFTemplateResolver
from ef_utils import create_aws_clients, get_autoscaling_group_properties, fail
from ef_conf_utils import pull_repo

# CONSTANTS
# Cloudformation template size limit in bytes (which translates to the length of the template)
CLOUDFORMATION_SIZE_LIMIT = 51200

class EFCFContext(EFContext):
  def __init__(self):
    super(EFCFContext, self).__init__()
    self._changeset = None
    self._lint = None
    self._poll_status = None
    self._template_file = None

  @property
  def changeset(self):
    """True if the tool should generate a changeset rather than executing the change immediately"""
    return self._changeset

  @changeset.setter
  def changeset(self, value):
    if type(value) is not bool:
      raise TypeError("changeset value must be bool")
    self._changeset = value

  @property
  def lint(self):
    """True if the tool should lint the rendered template rather than uploading to cloudformation"""
    return self._lint

  @lint.setter
  def lint(self, value):
    if type(value) is not bool:
      raise TypeError("lint value must be bool")
    self._lint = value

  @property
  def poll_status(self):
    """True if the tool should poll for stack status"""
    return self._poll_status

  @poll_status.setter
  def poll_status(self, value):
    if type(value) is not bool:
      raise TypeError("poll_status value must be bool")
    self._poll_status = value

  @property
  def template_file(self):
    """Path to the template file"""
    return self._template_file

  @template_file.setter
  def template_file(self, value):
    if type(value) is not str:
      raise TypeError("template file value must be str")
    self._template_file = value


def handle_args_and_set_context(args):
  """
  Args:
    args: the command line args, probably passed from main() as sys.argv[1:]
  Returns:
    a populated EFCFContext object (extends EFContext)
  Raises:
    IOError: if service registry file can't be found or can't be opened
    RuntimeError: if branch isn't as spec'd in ef_config.EF_REPO_BRANCH
    CalledProcessError: if 'git rev-parse' command to find repo root could not be run
  """
  parser = argparse.ArgumentParser(description="Render cloudformation templates, create changesets, or update "
                                               "cloudformation stacks in an AWS account.")
  parser.add_argument("template_file", help="/path/to/template_file.json")
  parser.add_argument("env", help=", ".join(EFConfig.ENV_LIST))
  parser.add_argument("--sr", help="optional /path/to/service_registry_file.json", default=None)
  parser.add_argument("--verbose", help="Print additional info + resolved template", action="store_true", default=False)
  parser.add_argument("--devel", help="Allow running from branch; don't refresh from origin", action="store_true",
                      default=False)
  group = parser.add_mutually_exclusive_group()
  group.add_argument("--changeset", help="create a changeset; cannot be combined with --commit",
                      action="store_true", default=False)
  group.add_argument("--commit", help="Make changes in AWS (dry run if omitted); cannot be combined with --changeset",
                      action="store_true", default=False)
  group.add_argument("--lint", help="Execute cfn-lint on the rendered template", action="store_true",
                      default=False)
  group.add_argument("--render", help="Output resolved template", action="store_true", default=False)
  parser.add_argument("--percent", help="Specifies an override to the percentage of instances in an Auto Scaling rolling update (e.g. 10 for 10%%)",
                      type=int, default=False)
  parser.add_argument("--poll", help="Poll Cloudformation to check status of stack creation/updates",
                      action="store_true", default=False)
  parser.add_argument("--skip_symbols", help="Skip resolving the provided symbols", nargs='+', default=[])
  parser.add_argument("--custom_rules", help="A directory with custom rules for cfn-lint")

  parsed_args = vars(parser.parse_args(args))
  context = EFCFContext()
  try:
    context.env = parsed_args["env"]
    context.template_file = parsed_args["template_file"]
  except ValueError as e:
    fail("Error in argument: {}".format(e))
  context.changeset = parsed_args["changeset"]
  context.commit = parsed_args["commit"]
  context.devel = parsed_args["devel"]
  context.lint = parsed_args["lint"]
  context.custom_rules = parsed_args["custom_rules"]
  context.percent = parsed_args["percent"]
  context.poll_status = parsed_args["poll"]
  context.skip_symbols = parsed_args["skip_symbols"]
  context.verbose = parsed_args["verbose"]
  context.render = parsed_args["render"]
  # Set up service registry and policy template path which depends on it
  context.service_registry = EFServiceRegistry(parsed_args["sr"])
  return context

def resolve_template(template, profile, env, region, service, skip_symbols, verbose):
  # resolve {{SYMBOLS}} in the passed template file
  os.path.isfile(template) or fail("Not a file: {}".format(template))
  resolver = EFTemplateResolver(profile=profile, target_other=True, env=env,
                                region=region, service=service, skip_symbols=skip_symbols, verbose=verbose)
  with open(template) as template_file:
    resolver.load(template_file)
    resolver.render()

  if verbose:
    print(resolver.template)

  dangling_left, dangling_right = resolver.count_braces()
  if resolver.unresolved_symbols():
    fail("Unable to resolve symbols: " + ",".join(["{{" + s + "}}" for s in resolver.unresolved_symbols()]))
  elif dangling_left > 0 or dangling_right > 0:
    fail("Some {{ or }} were not resolved. left{{: {}, right}}: {}".format(dangling_left, dangling_right))
  else:
    return resolver.template


def is_stack_termination_protected_env(env):
  return env in EFConfig.STACK_TERMINATION_PROTECTED_ENVS


def enable_stack_termination_protection(clients, stack_name):
  clients["cloudformation"].update_termination_protection(
    EnableTerminationProtection=True,
    StackName=stack_name
  )

def calculate_max_batch_size(asg_client, service, percent):
  autoscaling_group_properties = get_autoscaling_group_properties(asg_client, service.split("-")[0], "-".join(service.split("-")[1:]))
  if not autoscaling_group_properties:
      # safe default
      return 1
  current_desired = autoscaling_group_properties[0]["DesiredCapacity"]
  new_batch_size = int(math.ceil(current_desired * (percent * 0.01)))
  # max batch size must be a minimum of 1, otherwise cloudformation gives an error.
  return max(new_batch_size, 1)

class CFTemplateLinter(object):

  def __init__(self, template, custom_rules):
    self.template = template
    self.work_dir = '.lint'
    self.local_template_path = os.path.join(self.work_dir, 'template.json')
    self.custom_rules = custom_rules
    self.cfn_exit_code = None
    self.exit_code = None
    self.setup()

  def setup(self):
    if not os.path.exists(self.work_dir):
      os.mkdir(self.work_dir)

    with open(self.local_template_path, 'w') as f:
      f.write(self.template)

  def run_tests(self):
    self.cfn_lint()
    self.teardown()

  def cfn_lint(self):
    print("=== CLOUDFORMATION LINTING ===")
    cmd = [
      'cfn-lint',
      '--template {}'.format(self.local_template_path)
    ]

    if self.custom_rules:
      cmd.append('--append-rules {}'.format(self.custom_rules))

    cfn = subprocess.Popen(' '.join(cmd), shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = cfn.communicate()
    print(stdout, stderr)
    if cfn.returncode in [0, 4]:
      print("Template passed CFN linting")
    self.cfn_exit_code = cfn.returncode

  def teardown(self):
    try:
      os.remove(self.local_template_path)
      os.rmdir(self.work_dir)
    except OSError as e:
      print("WARNING: Unable to remove local workdir or test-copy of template")
      print(e)
    self.exit_code = 1 if self.cfn_exit_code not in [0, 4] else 0  # Ignore cfn-lint warnings


def main():
  context = handle_args_and_set_context(sys.argv[1:])

  if context.changeset:
    print("=== CHANGESET ===\nCreating changeset only. See AWS GUI for changeset\n=== CHANGESET ===")
  elif not context.commit and not context.render:
    print("=== DRY RUN ===\nValidation only. Use --commit to push template to CF\n=== DRY RUN ===")

  service_name = os.path.basename(os.path.splitext(context.template_file)[0])
  template_file_dir = os.path.dirname(context.template_file)
  # parameter file may not exist, but compute the name it would have if it did
  parameter_file_dir = template_file_dir + "/../parameters"
  parameter_file = parameter_file_dir + "/" + service_name + ".parameters." + context.env_full + ".json"

  # If running in EC2, use instance credentials (i.e. profile = None)
  # unless it's a non-EC2, which means that we use local
  # credentials with profile name in .aws/credentials == account alias name
  if context.whereami == "ec2":
    profile = None
  else:
    profile = context.account_alias

  # Get service registry and refresh repo if appropriate
  try:
    if not context.devel and context.whereami != 'jenkins':
      pull_repo()
  except Exception as error:
    fail("Error: ", error)

  # Service must exist in service registry
  if context.service_registry.service_record(service_name) is None:
    fail("service: {} not found in service registry: {}".format(service_name, context.service_registry.filespec))

  if not context.env_full in context.service_registry.valid_envs(service_name):
    fail("Invalid environment: {} for service_name: {}\nValid environments are: {}" \
         .format(context.env_full, service_name, ", ".join(context.service_registry.valid_envs(service_name))))

  if context.percent and (context.percent <= 0 or context.percent > 100):
    fail("Percent value cannot be less than or equal to 0 and greater than 100")

  # Set the region found in the service_registry. Default is EFConfig.DEFAULT_REGION if region key not found
  region = context.service_registry.service_region(service_name)

  if context.verbose:
    print("service_name: {}".format(service_name))
    print("env: {}".format(context.env))
    print("env_full: {}".format(context.env_full))
    print("env_short: {}".format(context.env_short))
    print("region: {}".format(region))
    print("template_file: {}".format(context.template_file))
    print("parameter_file: {}".format(parameter_file))
    if profile:
      print("profile: {}".format(profile))
    print("whereami: {}".format(context.whereami))
    print("service type: {}".format(context.service_registry.service_record(service_name)["type"]))

  template = resolve_template(
    template=context.template_file,
    profile=profile,
    env=context.env,
    region=region,
    service=service_name,
    skip_symbols=context.skip_symbols,
    verbose=context.verbose
  )

  if context.render:
    print(template)
    exit()

  # Create clients - if accessing by role, profile should be None
  try:
    clients = create_aws_clients(region, profile, "cloudformation", "autoscaling")
  except RuntimeError as error:
    fail("Exception creating clients in region {} with profile {}".format(region, profile), error)

  stack_name = context.env + "-" + service_name
  try:
    stack_exists = clients["cloudformation"].describe_stacks(StackName=stack_name)
  except botocore.exceptions.ClientError:
    stack_exists = False

  # Load parameters from file
  if os.path.isfile(parameter_file):
    parameters_template = resolve_template(
      template=parameter_file,
      profile=profile,
      env=context.env,
      region=region,
      service=service_name,
      skip_symbols=context.skip_symbols,
      verbose=context.verbose
    )
    try:
      parameters = json.loads(parameters_template)
    except ValueError as error:
      fail("JSON error in parameter file: {}".format(parameter_file, error))
  else:
    parameters = []

  if context.percent:
    print("Modifying deploy rate to {}%".format(context.percent))
    modify_template = json.loads(template)
    for key in modify_template["Resources"]:
      if modify_template["Resources"][key]["Type"] == "AWS::AutoScaling::AutoScalingGroup":
        if "UpdatePolicy" in modify_template["Resources"][key]:
          autoscaling_group = modify_template["Resources"][key]["Properties"]
          service = autoscaling_group["Tags"][0]["Value"]
          autoscaling_group_properties = get_autoscaling_group_properties(clients["autoscaling"], service.split("-")[0], "-".join(service.split("-")[1:]))
          new_max_batch_size = calculate_max_batch_size(clients["autoscaling"], service, context.percent)
          modify_template["Resources"][key]["UpdatePolicy"]["AutoScalingRollingUpdate"]["MaxBatchSize"] = new_max_batch_size
          current_desired = autoscaling_group_properties[0]["DesiredCapacity"] if autoscaling_group_properties else "missing"
          print("Service {} [current desired: {}, calculated max batch size: {}]".format(
                service, current_desired, new_max_batch_size))
    template = json.dumps(modify_template)

  # Detect if the template exceeds the maximum size that is allowed by Cloudformation
  if len(template) > CLOUDFORMATION_SIZE_LIMIT:
    # Compress the generated template by removing whitespaces
    print("Template exceeds the max allowed length that Cloudformation will accept. Compressing template...")
    print("Uncompressed size of template: {}".format(len(template)))
    unpacked = json.loads(template)
    template = json.dumps(unpacked, separators=(",", ":"))
    print("Compressed size of template: {}".format(len(template)))

  # Validate rendered template before trying the stack operation
  if context.verbose:
    print("Validating template")
  try:
    clients["cloudformation"].validate_template(TemplateBody=template)
    json.loads(template)  # Tests for valid JSON syntax, oddly not handled above
  except botocore.exceptions.ClientError as error:
    fail("Template did not pass validation", error)
  except ValueError as e:  # includes simplejson.decoder.JSONDecodeError
    fail('Failed to decode JSON', e)

  print("Template passed validation")

  # DO IT
  try:
    if context.changeset:
      print("Creating changeset: {}".format(stack_name))
      results = clients["cloudformation"].create_change_set(
        StackName=stack_name,
        TemplateBody=template,
        Parameters=parameters,
        Capabilities=['CAPABILITY_AUTO_EXPAND', 'CAPABILITY_IAM', 'CAPABILITY_NAMED_IAM'],
        ChangeSetName=stack_name,
        ClientToken=stack_name
      )
      if is_stack_termination_protected_env(context.env):
        enable_stack_termination_protection(clients, stack_name)
      results_ids = {key: value for key, value in results.iteritems()
                     if key in ('Id', 'StackId')}
      print("Changeset Info: {}".format(json.dumps(results_ids)))
    elif context.commit:
      if stack_exists:
        print("Updating stack: {}".format(stack_name))
        clients["cloudformation"].update_stack(
          StackName=stack_name,
          TemplateBody=template,
          Parameters=parameters,
          Capabilities=['CAPABILITY_AUTO_EXPAND', 'CAPABILITY_IAM', 'CAPABILITY_NAMED_IAM']
        )
        if is_stack_termination_protected_env(context.env):
          enable_stack_termination_protection(clients, stack_name)
      else:
        print("Creating stack: {}".format(stack_name))
        clients["cloudformation"].create_stack(
          StackName=stack_name,
          TemplateBody=template,
          Parameters=parameters,
          Capabilities=['CAPABILITY_AUTO_EXPAND', 'CAPABILITY_IAM', 'CAPABILITY_NAMED_IAM']
        )
        if is_stack_termination_protected_env(context.env):
          enable_stack_termination_protection(clients, stack_name)
      if context.poll_status:
        while True:
          stack_status = clients["cloudformation"].describe_stacks(StackName=stack_name)["Stacks"][0]["StackStatus"]
          if context.verbose:
            print("{}".format(stack_status))
          if stack_status.endswith('ROLLBACK_COMPLETE'):
            print("Stack went into rollback with status: {}".format(stack_status))
            sys.exit(1)
          elif re.match(r".*_COMPLETE(?!.)", stack_status) is not None:
            break
          elif re.match(r".*_FAILED(?!.)", stack_status) is not None:
            print("Stack failed with status: {}".format(stack_status))
            sys.exit(1)
          elif re.match(r".*_IN_PROGRESS(?!.)", stack_status) is not None:
            time.sleep(EFConfig.EF_CF_POLL_PERIOD)
    elif context.lint:
      tester = CFTemplateLinter(template, context.custom_rules)
      tester.run_tests()
      exit(tester.exit_code)

  except botocore.exceptions.ClientError as error:
    if error.response["Error"]["Message"] in "No updates are to be performed.":
      # Don't fail when there is no update to the stack
      print("No updates are to be performed.")
    else:
      fail("Error occurred when creating or updating stack", error)

if __name__ == "__main__":
  main()



"""
Copyright 2016-2017 Ellation, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import json
import subprocess
from collections import Counter
from os.path import isfile, normpath

from ef_config import EFConfig


class EFServiceRegistry(object):
  """
  Wraps interactions with the Service Registry
  Tries to find service registry in repo if service_registry_file is omitted in constructor

  Args:
    service_registry_file - /path/to/service_registry; default: EF_Config.DEFAULT_SERVICE_REGISTRY_FILE in repo root
  """

  def __init__(self, service_registry_file=None):
    """
    Args:
      service_registry_file: the file containing the service registry
    Raises:
      IOError: if file can't be found or can't be opened
      RuntimeError: if branch isn't as spec'd in ef_config.EF_REPO_BRANCH
      CalledProcessError: if 'git rev-parse' command to find repo root could not be run
    """
    # If a file wasn't provided, try to fetch the default
    if service_registry_file is None:
      repo_root = subprocess.check_output(["git", "rev-parse", "--show-toplevel"]).rstrip()
      self._service_registry_file = normpath("{}/{}".format(repo_root, EFConfig.DEFAULT_SERVICE_REGISTRY_FILE))
    else:
      self._service_registry_file = service_registry_file
    if not isfile(self._service_registry_file):
      raise IOError("Not a file: {}".format(self._service_registry_file))
    # Read the service registry
    service_registry_fh = open(self._service_registry_file, "r")
    try:
      self.service_registry_json = json.load(service_registry_fh)
    except ValueError:
      raise Exception("Malformed service registry file")
    # Validate service registry
    # 1. All service groups listed in EFConfig.SERVICE_GROUPS must be present
    for service_group in EFConfig.SERVICE_GROUPS:
      if service_group not in self.service_registry_json:
        raise RuntimeError("service registry: {} doesn't have '{}' service group listed in EFConfig".format(
                           self._service_registry_file, service_group))
    # 2. A service name must be unique and can only belong to one group
    service_counts = Counter()
    for service_group in EFConfig.SERVICE_GROUPS:
      service_counts.update(self.services(service_group).keys())
    for service_group in service_counts:
      if service_counts.get(service_group) > 1:
        raise RuntimeError("service name appears more than once in service registry: {}".format(service_group))

  @property
  def filespec(self):
    """
    Returns:
      path/to/service_registry_file
    """
    return self._service_registry_file

  def services(self, service_group=None):
    """
    Args:
      service_group: optional name of service group
    Returns:
      if service_group is omitted or None, flattened dict of all service records in the service registry
      if service_group is present, dict of service records in that group
    """
    # Specific service group requested
    if service_group is not None:
      if service_group not in EFConfig.SERVICE_GROUPS:
        raise RuntimeError("service registry: {} doesn't have '{}' section listed in EFConfig".format(
          self._service_registry_file, service_group))
      else:
        return self.service_registry_json[service_group]
    # Specific service group not requested - flatten and return all service records
    else:
      result = dict()
      for service_group in EFConfig.SERVICE_GROUPS:
        result.update(self.service_registry_json[service_group])
      return result

  def iter_services(self, service_group=None):
    """
    Args:
      service_group: optional name of service group
    Returns:
      if service_group is omitted or None, an Iterator over all flattened service records in the service registry
      if service_group is present, an Iterator over all service records in that group
    """
    if service_group is not None:
      if service_group not in EFConfig.SERVICE_GROUPS:
        raise RuntimeError("service registry: {} doesn't have '{}' section listed in EFConfig".format(
          self._service_registry_file, service_group))
      return iter(self.service_registry_json[service_group].items())
    else:
      return iter(self.services().items())

  def _expand_ephemeral_env_names(self, envlist):
    """
    Args:
      envlist: a list of environment names, which may be "ephemeral" names which get expanded to specific environment names
    Returns:
      List[String]: envlist, with ephemeral environment names expanded into specific environment names
    """
    result = []
    for service_env in envlist:
      if service_env not in EFConfig.PROTECTED_ENVS and service_env in EFConfig.EPHEMERAL_ENVS:
        result.extend((lambda env=service_env: [env + str(x) for x in range(EFConfig.EPHEMERAL_ENVS[env])])())
      else:
        result.append(service_env)
    return result

  def valid_envs(self, service_name):
    """
    Args:
      service_name: the name of the service in the service registry
    Returns:
      a list of strings - all the valid environments for 'service'
    Raises:
      RuntimeError if the service wasn't found
    """
    service_record = self.service_record(service_name)
    if service_record is None:
      raise RuntimeError("service registry doesn't have service: {}".format(service_name))

    # Return empty list if service has no "environments" section
    if "environments" not in service_record:
      return []
    # Otherwise gather up the envs
    return self._expand_ephemeral_env_names(service_record["environments"])

  def valid_auto_deploy_envs(self, service_name):
    """
    Args:
      service_name: the name of the service in the service registry
    Returns:
      List[String]|bool: A list of environments to which it is ok to auto-deploy the given service,
                         or a boolean flag allowing/denying all the environments in the 'environments'
                         service record field to be auto-deployed.
    Raises:
      RuntimeError if the service wasn't found

    Note that environments in the return are filtered for also being in the service record's
    'environments' field, so in order to qualify for autodeployment, an environment needs to be
    in both lists.
    """
    service_record = self.service_record(service_name)
    if service_record is None:
      raise RuntimeError("service registry doesn't have service: {}".format(service_name))

    if "auto_deploy_environments" not in service_record:
      return []

    # The list of valid deployment environments, for this service
    envs = self.valid_envs(service_name)

    #  If the auto_deploy_environments field is a boolean false, return [].
    if isinstance(service_record["auto_deploy_environments"], bool) and not service_record["auto_deploy_environments"]:
      return []
    elif isinstance(service_record["auto_deploy_environments"], bool):
      #  If the auto_deploy_environments field is a boolean true, return the full list of 'environments' for the service.
      return envs
    else:
      # filter the auto-deploy environments to only include environments that are also in the 'environments' list.
      auto_dep_envs = self._expand_ephemeral_env_names(service_record["auto_deploy_environments"])
      return list(set(auto_dep_envs).intersection(envs))

  def service_record(self, service_name):
    """
    Args:
      service_name: the name of the service in the service registry
    Returns:
      the entire service record from the service registry or None if the record was not found
    """
    if service_name not in self.services():
      return None
    return self.services()[service_name]

  def service_group(self, service_name):
    """
    Args:
      service_name: the name of the service in the service registry
    Returns:
      the name of the group the service is in, or None of the service was not found
    """
    for group in EFConfig.SERVICE_GROUPS:
      if service_name in self.services(group):
        return group
    return None

  def service_region(self, service_name):
    """
    Args:
      service_name: the name of the service in the service registry
    Returns:
      the region the service is in, or EFConfig.DEFAULT_REGION if the region was not found
    """
    if "region" not in self.services()[service_name]:
      return EFConfig.DEFAULT_REGION
    else:
      return self.services()[service_name]["region"]

  def version_keys(self):
    return self.service_registry_json["version_keys"]

  def allows_latest(self, version_key_name):
    """
    Does this version key allow 'latest' as an option (e.g. "latest AMI" makes sense and is allowed)
    Args:
      version_key_name: the version key to check for "allow_latest"
    Returns:
      True if the version key allows latest, False if it does not
    Raises:
      ValueError if the key was not found
    """
    if version_key_name not in self.version_keys():
      raise RuntimeError("service registry doesn't have a version key entry for: {}".format(version_key_name))
    if "allow_latest" not in self.version_keys()[version_key_name]:
      raise RuntimeError("service registry key {} doesn't have an 'allow_latest' value".format(
        version_key_name))
    return self.version_keys()[version_key_name]["allow_latest"]



#!/usr/bin/env python

from setuptools import setup

import versioneer


def readme():
    with open('README.md') as f:
        return f.read()


setup(
    name='ef-open',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    packages=[
        'efopen'
    ],
    install_requires=[
        'boto3',
        'click<=7.1.2',
        'PyYAML<=5.3.1',
        'cfn-lint',
        'requests<=2.25.1',
        'tenacity==7.0.0',
        'yamllint<=1.25.0'
    ],
    extras_require={
        'test': [
            'mock',
            'pylint',
        ]
    },
    entry_points={
        'console_scripts': [
            'ef-cf=efopen.ef_cf:main',
            'ef-cf-diff=efopen.ef_cf_diff:main',
            'ef-check-config=efopen.ef_check_config:main',
            'ef-generate=efopen.ef_generate:main',
            'ef-instanceinit=efopen.ef_instanceinit:main',
            'ef-password=efopen.ef_password:main',
            'ef-resolve-config=efopen.ef_resolve_config:main',
            'ef-version=efopen.ef_version:main'
        ],
    },
    url='https://github.com/crunchyroll/ef-open',
    license="Apache License 2.0",
    author='Ellation, Inc.',
    author_email='ops@ellation.com',
    description='CloudFormation Tools by Ellation',
    long_description=readme(),
    long_description_content_type='text/markdown'
)



#!/usr/bin/env python

"""
Utility to change entries in the version registry.
Expects these permissions (for the /current account/ only -- prod and non-prod are segregated)
  - GetObject (to retrieve a version)
  - GetObjectVersion (to retrieve a specific version of an object)
  - ListBucket (to retrieve objects including versions of objects
  - ListBucketVersions (which allows listing /Object/ versions)
  - PutObject (to update an object)

Syntax:
  ef-version <service> <key> <env> --get
  ef-version <service> <key> <env> --set <value> --commit
  ef-version <service> <key> <env> --set =prod --commit
  ef-version <service> <key> <env> --set =staging --commit
  ef-version <service> <key> <env> --set =latest --commit

Service registry must be reachable for --set; isn't needed for  --get*
If doing --set, must run from within the repo so app can auto-locate the service registry file

Known issues:
  Race condition if independent read/set occurs on a key before S3 update is fully consistent
"""

from __future__ import print_function
import argparse
from inspect import isfunction
import json
from operator import itemgetter
import sys
import urllib2

from botocore.exceptions import ClientError

from ef_config import EFConfig
from ef_context import EFContext
from ef_service_registry import EFServiceRegistry
from ef_utils import create_aws_clients, fail
from ef_conf_utils import pull_repo
from ef_version_resolver import EFVersionResolver

VERBOSE = False

class EFVersionContext(EFContext):
  def __init__(self):
    super(EFVersionContext, self).__init__()
    # core stuff
    self._build_number = None
    self._commit_hash = None
    self._force_env_full = None
    self._get = None
    self._history = None
    self._key = None
    self._location = None
    self._limit = None
    self._noprecheck = None
    self._rollback = None
    self._rollback_to = ""
    self._service_name = None # Cheating - we don't care about the full service record so don't use context.service
    self._show = None
    self._stable = None
    self._value = None
    self._versionresolver = None

  @property
  def build_number(self):
    """Externally defined build number assoicated with version entity"""
    return self._build_number

  @build_number.setter
  def build_number(self, value):
    """Setter provided because this is writeable from other than init() because --rollback alters it"""
    self._build_number = value

  @property
  def commit_hash(self):
    """Commit hash associated with version entity"""
    return self._commit_hash

  @commit_hash.setter
  def commit_hash(self, value):
    """Setter provided because this is writeable from other than init() because --rollback alters it"""
    self._commit_hash = value

  @property
  def get(self):
    """True if we are 'getting' the latest version value. --get was selected"""
    return self._get

  @EFContext.env.getter
  def env(self):
    """Returns env or overrides the env value with env full if --env_full was selected"""
    if self._force_env_full:
      return self.env_full
    return self._env

  @property
  def history(self):
    return self._history

  @property
  def key(self):
    """The key being manipulated"""
    return self._key

  @property
  def location(self):
    """Location (URL) where version is deployed and can be lookup up"""
    return self._location

  @location.setter
  def location(self, value):
    """Setter provided because this is writeable from other than init() because --rollback alters it"""
    self._location = value

  @property
  def limit(self):
    """Limit to number of records to return"""
    return self._limit

  @limit.setter
  def limit(self, value):
    """Limit is changed sometimes by internal calls"""
    self._limit = value

  @property
  def noprecheck(self):
    return self._noprecheck

  @property
  def rollback(self):
    return self._rollback

  @property
  def rollback_to(self):
    return self._rollback_to

  @property
  def service_name(self):
    """The service being queried or set (name only, since gets don't use the service registry"""
    return self._service_name

  @property
  def show(self):
    """True if --show was set"""
    return self._show

  @property
  def stable(self):
    return self._stable

  @stable.setter
  def stable(self, value):
    """Setter provided because this is writeable from other than init() because --rollback alters it"""
    self._stable = value

  @property
  def value(self):
    """The value read from the key (for a get) or to assign to the key (for a set)"""
    return self._value

  @value.setter
  def value(self, value):
    """Setter provided because this is writeable from other than init() because --rollback alters it"""
    self._value = value

  @property
  def versionresolver(self):
    """An instantiated ef_versionresolver"""
    return self._versionresolver


class Version(object):
  """
  Holds one 'object version' at a time, provides it in several formats
  """

  def __init__(self, object_version):
    self._value = object_version["Body"].read()
    self._version_id = object_version["VersionId"]
    self._last_modified = object_version["LastModified"].strftime("%Y-%m-%dT%H:%M:%S%Z")

    metadata = object_version["Metadata"]
    self._build_number = metadata.get(EFConfig.S3_VERSION_BUILDNUMBER_KEY,"")
    self._commit_hash = metadata.get(EFConfig.S3_VERSION_COMMITHASH_KEY,"")
    self._location = metadata.get(EFConfig.S3_VERSION_LOCATION_KEY,"")
    self._modified_by = metadata.get(EFConfig.S3_VERSION_MODIFIEDBY_KEY,"")
    self._status = metadata.get(EFConfig.S3_VERSION_STATUS_KEY,"")

  def __str__(self):
    return "{} {} {} {} {} {} {} {}".format(self._value, self._build_number, self._commit_hash, self._last_modified,
                                            self._modified_by, self._version_id, self._location, self._status)

  def __repr__(self):
    return str(self.to_json())

  def to_json(self):
    """
    called by VersionEncoder.default() when doing json.dumps() on the object
    the json materializes in reverse order from the order used here
    """
    return {
        "build_number": self._build_number,
        "commit_hash": self._commit_hash,
        "last_modified": self._last_modified,
        "location": self._location,
        "modified_by": self._modified_by,
        "status": self._status,
        "value": self._value,
        "version_id": self._version_id
    }

  @property
  def build_number(self):
    return self._build_number

  @property
  def commit_hash(self):
    return self._commit_hash

  @property
  def last_modified(self):
    return self._last_modified

  @property
  def location(self):
    return self._location

  @property
  def modified_by(self):
    return self._modified_by

  @property
  def status(self):
    return self._status

  @property
  def value(self):
    return self._value


class VersionEncoder(json.JSONEncoder):
  """
  provide a json encoder for the Version class to support json.dumps() output
  """

  def default(self, obj):
    if isinstance(obj, Version):
      return obj.to_json()
    return json.JSONEncoder.default(self, obj)


# Utilities
def handle_args_and_set_context(args):
  """
  Args:
    args: the command line args, probably passed from main() as sys.argv[1:]
  Returns:
    a populated EFVersionContext object
  """
  parser = argparse.ArgumentParser(description="Perform version tracking of a key for a service in the "
                                               "service_registry.json")
  parser.add_argument("service_name", help="name of the service")
  parser.add_argument("key", help="version key to look up for <service_name> such as 'ami-id' (list in EF_Config)")
  parser.add_argument("env", help=", ".join(EFConfig.ENV_LIST))
  group = parser.add_mutually_exclusive_group(required=True)
  group.add_argument("--get", help="get current version", action="store_true")
  group.add_argument("--set", help="set current version of <key> to <value> for <service_name>")
  group.add_argument("--rollback", help="set current version to most recent 'stable' version in history",
                     action="store_true")
  group.add_argument("--rollback-to", help="rollback current version to <ami-id> in history",
                     action="store", metavar='<ami-id>')
  group.add_argument("--history", help="Show version history for env/service/key", choices=['json', 'text'])
  group.add_argument("--show", help="Show keys and values. '*' allowed for <key> and <env>",
                     action="store_true", default=False)
  parser.add_argument("--build",
                      help="On --set, also set the externally defined build number associated with the version entity",
                      default="")
  parser.add_argument("--commit_hash", help="On --set, also set the commit hash associated with the version entity",
                      default="")
  parser.add_argument("--commit", help="Actually --set or --rollback (dry run if omitted)",
                      action="store_true", default=False)
  parser.add_argument("--devel", help="Allow running from branch; don't refresh from origin", action="store_true",
                      default=False)
  parser.add_argument("--force_env_full", help="Override env with env_full for account-scoped environments",
                      action="store_true", default=False)
  parser.add_argument("--limit", help="Limit 'history', 'rollback', 'show' to first N records (default 100, max 1000)",
                      type=int, default=100)
  parser.add_argument("--location", help="On --set, also mark the url location of the static build's version file to"
                      "support dist-hash precheck", default="")
  parser.add_argument("--noprecheck", help="Flag is deprecated but left behind to not break things", action="store_true")
  parser.add_argument("--sr", help="optional /path/to/service_registry_file.json", default=None)
  parser.add_argument("--stable", help="On --set, also mark the version 'stable'", action="store_true")
  parser.add_argument("--verbose", help="Print additional info", action="store_true", default=False)
  # parse
  parsed_args = vars(parser.parse_args(args))
  context = EFVersionContext()
  # marshall the inherited context values
  context._build_number = parsed_args["build"]
  context._commit_hash = parsed_args["commit_hash"]
  context.commit = parsed_args["commit"]
  context.devel = parsed_args["devel"]
  context._force_env_full = parsed_args["force_env_full"]
  try:
    context.env = parsed_args["env"]
  except ValueError as e:
    fail("Error in env: {}".format(e))
  # marshall this module's additional context values
  context._get = parsed_args["get"]
  context._history = parsed_args["history"]
  context._key = parsed_args["key"]
  context._noprecheck = True
  if not 1 <= parsed_args["limit"] <= 1000:
    fail("Error in --limit. Valid range: 1..1000")
  context._limit = parsed_args["limit"]
  context._location = parsed_args["location"]
  context._rollback = parsed_args["rollback"]
  context._rollback_to = parsed_args["rollback_to"]
  context._service_name = parsed_args["service_name"]
  context._show = parsed_args["show"]
  context._stable = parsed_args["stable"]
  context._value = parsed_args["set"]
  # Set up service registry and policy template path which depends on it
  context.service_registry = EFServiceRegistry(parsed_args["sr"])

  # VERBOSE is global
  global VERBOSE
  VERBOSE = parsed_args["verbose"]

  validate_context(context)
  return context


def print_if_verbose(message):
  if VERBOSE:
    print(message, file=sys.stderr)


def validate_context(context):
  """
    Validate the context. Fails the process on an invalid context
    Args:
      context: a populated EFVersionContext object
  """

  # Key must be valid
  key_data = EFConfig.VERSION_KEYS.get(context.key)
  if not key_data:
    fail("invalid key: {}; see VERSION_KEYS in ef_config for supported keys".format(context.key))

  registry = context.service_registry
  service = registry.service_record(context.service_name)
  # Service must exist in service registry
  if not service:
    fail("service: {} not found in service registry: {}".format(
         context.service_name, registry.filespec))

  # Lookup allowed key for service type
  service_type = service["type"]
  allowed_types = key_data.get("allowed_types", [])
  if service_type not in allowed_types:
    fail("service_type: {} is not allowed for key {}; see VERSION_KEYS[KEY]['allowed_types']"
         "in ef_config and validate service registry entry".format(service_type, context.key))

  return True


def _get_stable_versions(context):
  """
  Get all stable versions
  Args:
    context: a populated EFVersionContext object

  Returns:
    List of Version objects representing stable versions in this service's history
  """
  versions = get_versions(context)
  stable_versions = []
  for version in versions:
    if version.status == EFConfig.S3_VERSION_STATUS_STABLE:
      stable_versions.append(version)
  return stable_versions


def _get_latest_version(context):
  """
  Get latest version in history of service
  Args:
    context: a populated EFVersionContext object

  Returns:
    A Version object representing the latest version in the service's history
  """
  original_limit = context.limit
  context.limit = 1
  latest_version = get_versions(context)
  context.limit = original_limit
  return latest_version[0] if len(latest_version) > 0 else None


def get_versions(context):
  """
  Get all versions of a key
  Args:
    context: a populated EFVersionContext object
    return_stable: (default:False) If True, stop fetching if 'stable' version is found; return only that version
  Returns:
    List of Version objects sorted in reverse by last_modified (newest version is first).
  """
  s3_key = "{}/{}/{}".format(context.service_name, context.env, context.key)
  object_version_list = context.aws_client("s3").list_object_versions(
      Bucket=EFConfig.S3_VERSION_BUCKET,
      Delimiter='/',
      MaxKeys=context.limit,
      Prefix=s3_key
  )

  if "Versions" not in object_version_list:
    return []

  object_versions = []
  for version in object_version_list["Versions"]:
    object_version = Version(context.aws_client("s3").get_object(
        Bucket=EFConfig.S3_VERSION_BUCKET,
        Key=s3_key,
        VersionId=version["VersionId"]
    ))
    object_versions.append(object_version)

  return sorted(object_versions, key=lambda v: v.last_modified, reverse=True)


def cmd_get(context):
  obj_value = context.versionresolver.lookup("{},{}/{}".format(context.key, context.env, context.service_name))
  print(obj_value)


def cmd_history(context):
  versions = get_versions(context)
  if context.history == "text":
    print("{}-{} {}".format(context.env, context.service_name, context.key))
    for v in versions:
      print(v)
  elif context.history == "json":
    print(json.dumps(versions, cls=VersionEncoder))


def get_version_by_value(context, value):
  """
  Get the latest version that matches the provided ami-id
  Args:
    context: a populated EFVersionContext object
    value: the value of the version to look for
  """
  versions = get_versions(context)
  for version in versions:
    if version.value == value:
      return version
  fail("Didn't find a matching version for: "
       "{}:{} in env/service: {}/{}".format(
          context.key, value,
          context.env, context.service_name))


def cmd_rollback(context):
  """
  Roll back by finding the most recent "stable" tagged version, and putting it again, so that
  it's the new "current" version.
  Args:
    context: a populated EFVersionContext object
  """
  stable_versions = _get_stable_versions(context)
  latest_version = _get_latest_version(context)
  for version in stable_versions:
    if latest_version and (version.value != latest_version.value):
      context.value = version.value
      context.commit_hash = version.commit_hash
      context.build_number = version.build_number
      context.location = version.location
      context.stable = True
      cmd_set(context)
      return
  fail("Didn't find a version marked stable for key: {} in env/service: {}/{}".format(context.key, context.env,
                                                                                      context.service_name))


def cmd_rollback_to(context):
  """
  Roll back by finding a specific version in the history of the service and
  putting it as the new current version.
  Args:
    context: a populated EFVersionContext object
  """
  version = get_version_by_value(context, context.rollback_to)
  context.value = version.value
  context.commit_hash = version.commit_hash
  context.build_number = version.build_number
  context.location = version.location
  context.stable = True
  cmd_set(context)


def _get_deployed_ami_id(context):
  """
  NOTE: Currently not used until the future
  Get the most recent AMI ID for a service
  Args:
    context: a populated EFVersionContext object
  Returns:
    ImageId or None if no images exist or on error
  """
  try:
    response = context.aws_client("ec2").describe_images(
        Filters=[
            {"Name": "is-public", "Values": ["false"]},
            {"Name": "name", "Values": [context.service_name + EFConfig.AMI_SUFFIX + "*"]}
        ])
  except:
    return None
  if len(response["Images"]) > 0:
    return sorted(response["Images"], key=itemgetter('CreationDate'), reverse=True)[0]["ImageId"]
  else:
    return None


def cmd_set(context):
  """
  Set the new "current" value for a key.
  If the existing current version and the new version have identical /value/ and /status,
   then nothing is written, to avoid stacking up redundant entreis in the version table.
  Args:
    context: a populated EFVersionContext object
  """
  # If key value is a special symbol, see if this env allows it
  if context.value in EFConfig.SPECIAL_VERSIONS and context.env_short not in EFConfig.SPECIAL_VERSION_ENVS:
    fail("special version: {} not allowed in env: {}".format(context.value, context.env_short))
  # If key value is a special symbol, the record cannot be marked "stable"
  if context.value in EFConfig.SPECIAL_VERSIONS and context.stable:
    fail("special versions such as: {} cannot be marked 'stable'".format(context.value))

  # Resolve any references
  if context.value == "=prod":
    context.value = context.versionresolver.lookup("{},{}/{}".format(context.key, "prod", context.service_name))
  elif context.value == "=staging":
    context.value = context.versionresolver.lookup("{},{}/{}".format(context.key, "staging", context.service_name))
  elif context.value == "=latest":
    if not EFConfig.VERSION_KEYS[context.key]["allow_latest"]:
      fail("=latest cannot be used with key: {}".format(context.key))
    func_name = "_getlatest_" + context.key.replace("-", "_")
    if func_name in globals() and isfunction(globals()[func_name]):
      context.value = globals()[func_name](context)
    else:
      raise RuntimeError("{} version for {}/{} is '=latest' but can't look up because method not found: {}".format(
                         context.key, context.env, context.service_name, func_name))

  s3_key = "{}/{}/{}".format(context.service_name, context.env, context.key)
  s3_version_status = EFConfig.S3_VERSION_STATUS_STABLE if context.stable else EFConfig.S3_VERSION_STATUS_UNDEFINED

  # If the set would put a value and status that are the same as the existing 'current' value/status, don't do it
  context.limit = 1
  current_version = get_versions(context)
  # If there is no 'current version' it's ok, just means the set will write the first entry
  if len(current_version) == 1 and current_version[0].status == s3_version_status and \
          current_version[0].value == context.value:
      print("Version not written because current version and new version have identical value and status: {} {}"
            .format(current_version[0].value, current_version[0].status))
      return

  if not context.commit:
    print("=== DRY RUN ===\nUse --commit to set value\n=== DRY RUN ===")
    print("would set key: {} with value: {} {} {} {} {}".format(
          s3_key, context.value, context.build_number, context.commit_hash, context.location, s3_version_status))
  else:
    context.aws_client("s3").put_object(
        ACL='bucket-owner-full-control',
        Body=context.value,
        Bucket=EFConfig.S3_VERSION_BUCKET,
        ContentEncoding=EFConfig.S3_VERSION_CONTENT_ENCODING,
        Key=s3_key,
        Metadata={
            EFConfig.S3_VERSION_BUILDNUMBER_KEY: context.build_number,
            EFConfig.S3_VERSION_COMMITHASH_KEY: context.commit_hash,
            EFConfig.S3_VERSION_LOCATION_KEY: context.location,
            EFConfig.S3_VERSION_MODIFIEDBY_KEY: context.aws_client("sts").get_caller_identity()["Arn"],
            EFConfig.S3_VERSION_STATUS_KEY: s3_version_status
        },
        StorageClass='STANDARD'
    )
    print("set key: {} with value: {} {} {} {} {}".format(
          s3_key, context.value, context.build_number, context.commit_hash, context.location, s3_version_status))


def cmd_show(context):
  print("cmd_show is not implemented")


def main():
  # Fetch args and load context
  context = handle_args_and_set_context(sys.argv[1:])

  # Refresh from repo if necessary and possible (gets don't need service registry, sets do)
  if (context.rollback or context.value) and (not context.devel and context.whereami != 'jenkins'):
    print("Refreshing repo")
    try:
      pull_repo()
    except RuntimeError as error:
      fail("Error checking or pulling repo", error)

  # Sign on to AWS and create clients
  if context.whereami in ["ec2"]:
    # Always use instance credentials in EC2. One day we'll have "lambda" in there too, so use "in" w/ list
    aws_session_alias = None
  else:
    # Otherwise use local user credential matching the account alias
    aws_session_alias = context.account_alias
  # Make AWS clients
  try:
    context.set_aws_clients(create_aws_clients(EFConfig.DEFAULT_REGION, aws_session_alias, "ec2", "s3", "sts"))
  except RuntimeError:
    fail("Exception creating AWS client in region {} with aws account alias {} (None=instance credentials)".format(
         EFConfig.DEFAULT_REGION, aws_session_alias))

  # Instantiate a versionresolver - we'll use some of its methods
  context._versionresolver = EFVersionResolver(context.aws_client())

  # Carry out the requested action
  if context.get:
    cmd_get(context)
  elif context.history:
    cmd_history(context)
  elif context.rollback:
    cmd_rollback(context)
  elif context.rollback_to:
    cmd_rollback_to(context)
  elif context.show:
    cmd_show(context)
  elif context.value:
    cmd_set(context)


if __name__ == "__main__":
  main()



"""
Copyright 2016-2017 Ellation, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from os import walk
import yaml

import botocore.exceptions

from ef_config import EFConfig
import ef_utils
import ef_conf_utils


class EFInstanceinitConfigReader:
  """
  Reads a set of local configs files and parameters, stored locally as files or in s3
  Config_Reader.next() advances to the next item, or returns False if there are no more
  Args:
    service: name of the service
    base_path: path to TLD of all configs that should be loaded
  """

  def __init__(self, service, base_path, logger, s3_resource=None):
    """
    Args:
      service: "s3" or "file" -- use S3 in AWS, and "file" in local dev
      base_path: prefix of the path to the config bucket:
        for s3: "all" or "<service>"
        for file: "/vagrant/configs/<"all"|<service>>
      logger: an function that will log messages from here
      s3_resource: if s3, a boto3 s3 resource
    """
    self.service = service
    self.base_path = base_path
    self.logger = logger
    self.s3_resource = s3_resource
    self.template_prefix = self.base_path + "/templates/"
    self.current = None

    self.items = []
    if self.service == "s3":
      bucket = self.s3_resource.Bucket(EFConfig.S3_CONFIG_BUCKET)
      bucket_objects = bucket.objects.filter(Prefix=self.template_prefix)
      # unpack into a list for easier iteration
      for bucket_object in bucket_objects:
        self.items.append(bucket_object)
    elif self.service == "file":
      for _path, _dir, _files in walk(self.template_prefix):
        for _file in _files:
          self.items.append("{}{}".format(_path, _file))
    else:
      raise ValueError("invalid service: {} in EFInstanceinitConfigReader; valid: 's3','file'".format(self.service))

  def next(self):
    if len(self.items) <= 0:
      return False
    self.current = self.items.pop()
    self.logger("next item: " + repr(self.current))
    # S3
    if self.service == "s3":
      # the S3 set contains path entries with length 0; skip them
      while self.current.size == 0:
        if len(self.items) < 1:
          return False
        self.current = self.items.pop()
    return True

  @property
  def template(self):
    if self.service == "s3":
      try:
        self.logger("Loading template object: " + self.current.key)
        return self.current.get()['Body'].read()
      except botocore.exceptions.ClientError as error:
        raise IOError("Error loading template at key: {} {}".format(self.current.key, repr(error)))
    elif self.service == "file":
      self.logger("Loading template file: {}".format(self.current))
      try:
        return open(self.current).read()
      except Exception as error:
        raise IOError("Error loading template file: {} {}".format(self.current, repr(error)))

  @property
  def parameters(self):
    if self.service == "s3":
      key = ef_conf_utils.get_template_parameters_s3(self.current.key, self.s3_resource)
      self.logger("Loading parameters object: {}".format(key))
      try:
        obj = self.s3_resource.Object(EFConfig.S3_CONFIG_BUCKET, key)
        body = obj.get()['Body'].read()
        return yaml.safe_load(body)
      except botocore.exceptions.ClientError as error:
        raise IOError("Error loading parameters from: {} {}".format(key, repr(error)))
    elif self.service == "file":
      parameters_file = ef_conf_utils.get_template_parameters_file(self.current)
      self.logger("Loading parameters file: {}".format(self.current))
      try:
        return yaml.safe_load(file(parameters_file))
      except Exception as error:
        raise IOError("Error loading parameters file: {} {}".format(parameters_file, repr(error)))

  @property
  def dest(self):
    return self.parameters["dest"]

  @property
  def current_key(self):
    if self.service == "s3":
      return self.current.key
    elif self.service == "file":
      return self.current



"""
Copyright 2016-2017 Ellation, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import yaml

import botocore.exceptions

from ef_config import EFConfig
from ef_conf_utils import get_env_short


class EFAppConfigReader:
  """
  Reads one set of env-based configs (one object) directly from a S3 for use by an app (no file overlaying).
  Perform hierarchical env-based config overlaying (least-specific to most)
  e.g.: default (if present) - superseded by proto (if present) - superseded by proto<N> (if present)
  or: default (if present) - superseded by staging or prod (if present)
  """

  def __init__(self, env, service, clients):
    """
    Args:
      env: the environment to resolve values for
      service: "all" or "<service">
      clients: a dict of AWS clients, with at least "s3"
    Raises:
      botocore.exceptions.ClientError if config
    """
    self.clients = clients
    self.env = env
    self.env_short = get_env_short(env)
    self.parameters = None
    self.service = service

    # Load from S3
    parameters_key = self.service + "/parameters/" + self.service + ".parameters.json"
    try:
      parameters_object = self.clients["s3"].get_object(
        Bucket=EFConfig.S3_CONFIG_BUCKET,
        Key=parameters_key
      )
    except botocore.exceptions.ClientError:
      raise RuntimeError("Error getting parameters from key: {}".format(parameters_key))
    self.parameters = yaml.safe_load(parameters_object["Body"].read().decode("utf-8"))["params"]

  def __repr__(self):
    keys = set()
    for env in self.parameters.keys():
      keys.update(self.parameters[env].keys())
    result = ""
    for key in keys:
      result = result + "{}: {}\n".format(key, self.get_value(key))
    return result

  def get_value(self, symbol):
    """
    Hierarchically searches for 'symbol' in the parameters blob if there is one (would have
    been retrieved by 'load()'). Order is: default, <env_short>, <env>
    Args:
      symbol: the key to resolve
    Returns:
      Hierarchically resolved value for 'symbol' in the environment set by the constructor,
      or None if a match is not found or there are no parameters
    """
    default = "default"
    if not self.parameters:
      return None
    # Hierarchically lookup the value
    result = None
    if default in self.parameters and symbol in self.parameters[default]:
      result = self.parameters[default][symbol]
    if self.env_short in self.parameters and symbol in self.parameters[self.env_short]:
      result = self.parameters[self.env_short][symbol]
    # This lookup is redundant when env_short == env, but it's also cheap
    if self.env in self.parameters and symbol in self.parameters[self.env]:
      result = self.parameters[self.env][symbol]
    # Finally, convert any list of items into a single \n-delimited string
    if isinstance(result, list):
      result = "\n".join(result)

    return result


# @todo: proper test coverage for this module
# In lieu of that for the moment, this demonstration code:
import boto3


def main():
  session = boto3.Session(profile_name="ellationeng", region_name="us-west-2")
  clients = {"s3": session.client("s3")}
  configreader = EFAppConfigReader("staging", "qc-pulls", clients)
  print("remote username")
  print(configreader.get_value("remote_username"))


if __name__ == "__main__":
  main()



#!/usr/bin/env python

"""
Manual single file config resolver

This is mostly for testing and teaching - accepts a config file as input, finds
the matching config blob in /configs, resolves everything, and outputs the result
as it would be written to a file on an instance when starting up

Copyright 2016-2017 Ellation, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import print_function

import argparse
import json
import subprocess
import sys
from os.path import abspath, dirname, normpath

import yaml
from yamllint import linter as yamllinter
from yamllint import config as yamllint_config

import ef_utils
import ef_conf_utils
from ef_config import EFConfig
from ef_template_resolver import EFTemplateResolver
from ef_utils import fail
from ef_conf_utils import get_account_alias


class Context:
  def __init__(self, profile, region, env, service, template_path, no_params, verbose, lint, silent, skip_symbols):
    self.profile = profile
    self.region = region
    self.env = env
    self.service = service
    self.template_path = template_path
    self.no_params = no_params
    self.param_path = ef_conf_utils.get_template_parameters_file(self.template_path)
    self.verbose = verbose
    self.lint = lint
    self.silent = silent
    self.skip_symbols = skip_symbols

  def __str__(self):
    return("profile: {}\nregion: {}\nenv: {}\nservice: {}\ntemplate_path: {}\nparam_path: {}\nlint: {}".format(
      self.profile, self.region, self.env, self.service, self.template_path, self.param_path, self.lint))


def handle_args_and_set_context(args):
  """
  Args:
    args: the command line args, probably passed from main() as sys.argv[1:]
  Returns:
    a populated Context object based on CLI args
  """
  parser = argparse.ArgumentParser()
  parser.add_argument("env", help="environment")
  parser.add_argument("path_to_template", help="path to the config template to process")
  parser.add_argument("--no_params", help="disable loading values from params file", action="store_true", default=False)
  parser.add_argument("--verbose", help="Output extra info", action="store_true", default=False)
  parser.add_argument("--lint", help="Test configs for valid JSON/YAML syntax", action="store_true", default=False)
  parser.add_argument("--silent", help="Suppress output of rendered template", action="store_true", default=False)
  parser.add_argument("--skip_symbols", help="Skip resolving the provided symbols", nargs='+', default=[])
  parsed = vars(parser.parse_args(args))
  path_to_template = abspath(parsed["path_to_template"])
  service = path_to_template.split('/')[-3]

  return Context(
      get_account_alias(parsed["env"]),
      EFConfig.DEFAULT_REGION,
      parsed["env"],
      service,
      path_to_template,
      parsed["no_params"],
      parsed["verbose"],
      parsed["lint"],
      parsed["silent"],
      parsed["skip_symbols"]
  )


def merge_files(context):
  """
  Given a context containing path to template, env, and service:
  merge config into template and output the result to stdout
  Args:
    context: a populated context object
  """
  resolver = EFTemplateResolver(
      profile=context.profile,
      region=context.region,
      env=context.env,
      service=context.service,
      skip_symbols=context.skip_symbols
  )

  try:
    with open(context.template_path, 'r') as f:
      template_body = f.read()
      f.close()
  except IOError as error:
    raise IOError("Error loading template file: {} {}".format(context.template_path, repr(error)))

  if context.no_params is False:
    try:
      with open(context.param_path, 'r') as f:
        param_body = f.read()
        f.close()
    except IOError as error:
      raise IOError("Error loading param file: {} {}".format(context.param_path, repr(error)))

    dest = yaml.safe_load(param_body)["dest"]

    # if 'dest' for the current object contains an 'environments' list, check it
    if "environments" in dest:
      if not resolver.resolved["ENV_SHORT"] in dest["environments"]:
        print("Environment: {} not enabled for {}".format(resolver.resolved["ENV_SHORT"], context.template_path))
        return

    # Process the template_body - apply context + parameters
    resolver.load(template_body, param_body)
  else:
    resolver.load(template_body)
  rendered_body = resolver.render()

  if not resolver.resolved_ok():
    raise RuntimeError("Couldn't resolve all symbols; template has leftover {{ or }}: {}".format(resolver.unresolved_symbols()))

  if context.lint:
    if context.template_path.endswith(".json"):
      try:
        json.loads(rendered_body, strict=False)
        print("JSON passed linting process.")
      except ValueError as e:
        fail("JSON failed linting process.", e)
    elif context.template_path.endswith((".yml", ".yaml")):
      conf = yamllint_config.YamlLintConfig(content='extends: relaxed')
      lint_output = yamllinter.run(rendered_body, conf)
      lint_level = 'error'
      lint_errors = [issue for issue in lint_output if issue.level == lint_level]
      if lint_errors:
        split_body = rendered_body.splitlines()
        for error in lint_errors:
          print(error)
          # printing line - 1 because lists start at 0, but files at 1
          print("\t", split_body[error.line - 1])
        fail("YAML failed linting process.")

  if context.verbose:
    print(context)
    if context.no_params:
      print('no_params flag set to true!')
      print('Inline template resolution based on external symbol lookup only and no destination for file write.\n')
    else:
      dir_path = normpath(dirname(dest["path"]))
      print("make directories: {} {}".format(dir_path, dest["dir_perm"]))
      print("chmod file to: " + dest["file_perm"])
      user, group = dest["user_group"].split(":")
      print("chown last directory in path to user: {}, group: {}".format(user, group))
      print("chown file to user: {}, group: {}\n".format(user, group))

    print("template body:\n{}\nrendered body:\n{}\n".format(template_body, rendered_body))
  elif context.silent:
    print("Config template rendered successfully.")
  else:
    print(rendered_body)


def main():
  context = handle_args_and_set_context(sys.argv[1:])
  merge_files(context)

if __name__ == "__main__":
  main()



import json
import logging

import requests
from tenacity import (
  retry,
  retry_if_exception,
  stop_after_attempt,
  wait_exponential,
)

logger = logging.getLogger(__name__)


def check_5xx_exception(ex):
  return isinstance(ex, requests.exceptions.HTTPError) and ex.response.status_code > 500

exponential_retry_on_5xx = retry(retry=retry_if_exception(check_5xx_exception), stop=stop_after_attempt(5), wait=wait_exponential(min=2, max=10))

class AlertPolicy(object):

  def __init__(self, env, service):
    self._env = env
    self._service = service
    self._id = None
    self.name = None
    self.notification_channels = None
    self.config_conditions = None
    self.remote_conditions = None
    self.local_alert_nrql_conditions = None
    self.remote_alert_nrql_conditions = None
    self.local_alert_apm_conditions = None
    self.remote_alert_apm_conditions = None
    self.symbols = None
    self.set_name()
    self.set_symbols()

  @property
  def id(self):
    return self._id

  @id.setter
  def id(self, value):
    try:
      self._id = int(value)
      self.set_symbols()
    except ValueError:
      logger.error("Invalid value '{}' for policy id.".format(value))

  @property
  def env(self):
    return self._env

  @env.setter
  def env(self, value):
    if isinstance(value, str):
      self._env = value
      self.set_name()
      self.set_symbols()
    else:
      logger.error("Invalid value '{}' for env.".format(value))

  @property
  def service(self):
    return self._service

  @service.setter
  def service(self, value):
    if isinstance(value, str):
      self._env = value
      self.set_name()
      self.set_symbols()
    else:
      logger.error("Invalid value '{}' for service.".format(value))

  def set_name(self):
    self.name = "{}-{}".format(self.env, self.service)

  def set_symbols(self):
    self.symbols = {
      "ENV": self.env,
      "POLICY_ID": self.id,
      "SERVICE": self.service
    }


class NewRelic(object):

  def __init__(self, admin_token):
    self.admin_token = admin_token
    self.auth_header = {'X-Api-Key': self.admin_token, 'Content-Type': 'application/json'}
    self.all_alert_policies = None
    self.all_notification_channels = None
    self.refresh_alert_policies()
    self.refresh_notification_channels()

  @exponential_retry_on_5xx
  def iterative_get(self, endpoint_url, response_key, params=None):
    params = params if params else {}
    response = []
    while True:
      r = requests.get(
        url=endpoint_url,
        headers=self.auth_header,
        params=params
      )
      self._log_response_errors(r)
      r.raise_for_status()
      response.extend(r.json()[response_key])
      try:
        endpoint_url = r.links['next']['url']
      except KeyError:
        # This parameter comes only on GET https://infra-api.newrelic.com/v2/alerts/conditions
        if 'meta' not in r.json():
          break

        meta = r.json()['meta']

        new_offset = meta['offset'] + meta['limit']
        if new_offset < meta['total']:
          params['offset'] = new_offset
        else:
          break
    return response

  def refresh_alert_policies(self):
    self.all_alert_policies = self.iterative_get(
      endpoint_url='https://api.newrelic.com/v2/alerts_policies.json',
      response_key='policies'
    )

  def refresh_notification_channels(self):
    self.all_notification_channels = self.iterative_get(
      endpoint_url='https://api.newrelic.com/v2/alerts_channels.json',
      response_key='channels'
    )

  def alert_policy_exists(self, policy_name):
    """Check to see if an alert policy exists in NewRelic. Return True if so, False if not"""
    if next((policy for policy in self.all_alert_policies if policy['name'] == policy_name), False):
      return True

  @exponential_retry_on_5xx
  def create_alert_policy(self, policy_name):
    """Creates an alert policy in NewRelic"""
    policy_data = { 'policy': { 'incident_preference': 'PER_POLICY', 'name': policy_name } }
    create_policy = requests.post(
      'https://api.newrelic.com/v2/alerts_policies.json',
      headers=self.auth_header,
      data=json.dumps(policy_data))
    self._log_response_errors(create_policy)
    create_policy.raise_for_status()
    policy_id = create_policy.json()['policy']['id']
    self.refresh_alert_policies()
    return policy_id

  @exponential_retry_on_5xx
  def add_policy_channels(self, policy_id, channel_ids):
    payload = { 'policy_id': policy_id, 'channel_ids': channel_ids}
    put_channels = requests.put(
      url='https://api.newrelic.com/v2/alerts_policy_channels.json',
      headers=self.auth_header,
      params=payload
    )
    self._log_response_errors(put_channels)
    put_channels.raise_for_status()
    return

  @exponential_retry_on_5xx
  def delete_policy_channel(self, policy_id, channel_id):
    payload = {'policy_id': policy_id, 'channel_id': channel_id}
    delete_channel = requests.delete(
      url='https://api.newrelic.com/v2/alerts_policy_channels.json',
      headers=self.auth_header,
      params=payload
    )
    self._log_response_errors(delete_channel)
    delete_channel.raise_for_status()
    return

  @exponential_retry_on_5xx
  def create_alert_condition(self, condition):
    add_condition = requests.post(
      url='https://infra-api.newrelic.com/v2/alerts/conditions',
      headers=self.auth_header,
      data=json.dumps({"data": condition})
    )

    self._log_response_errors(add_condition)
    add_condition.raise_for_status()
    return

  def get_policy_alert_conditions(self, policy_id):
    return self.iterative_get(
      endpoint_url='https://infra-api.newrelic.com/v2/alerts/conditions',
      response_key='data',
      params={'policy_id': policy_id}
    )

  @exponential_retry_on_5xx
  def delete_policy_alert_condition(self, condition_id):
    delete_condition = requests.delete(
      url='https://infra-api.newrelic.com/v2/alerts/conditions/{}'.format(condition_id),
      headers=self.auth_header
    )
    self._log_response_errors(delete_condition)
    delete_condition.raise_for_status()
    return

  @exponential_retry_on_5xx
  def create_alert_nrql_condition(self, policy_id, condition):
    add_condition = requests.post(
      url='https://api.newrelic.com/v2/alerts_nrql_conditions/policies/{}.json'.format(policy_id),
      headers=self.auth_header,
      data=json.dumps({"nrql_condition": condition})
    )

    self._log_response_errors(add_condition)
    add_condition.raise_for_status()
    return

  def get_policy_alert_nrql_conditions(self, policy_id):
    return self.iterative_get(
      endpoint_url='https://api.newrelic.com/v2/alerts_nrql_conditions.json',
      response_key='nrql_conditions',
      params={'policy_id': policy_id}
    )

  @exponential_retry_on_5xx
  def put_policy_alert_nrql_condition(self, condition_id, condition):
    put_condition = requests.put(
      url='https://api.newrelic.com/v2/alerts_nrql_conditions/{}.json'.format(condition_id),
      headers=self.auth_header,
      data=json.dumps({"nrql_condition": condition})
    )

    self._log_response_errors(put_condition)
    put_condition.raise_for_status()
    return


  @exponential_retry_on_5xx
  def create_alert_apm_condition(self, policy_id, condition):
    add_condition = requests.post(
      url='https://api.newrelic.com/v2/alerts_conditions/policies/{}.json'.format(policy_id),
      headers=self.auth_header,
      data=json.dumps({"condition": condition})
    )

    self._log_response_errors(add_condition)
    add_condition.raise_for_status()
    return

  def get_policy_alert_apm_conditions(self, policy_id):
    return self.iterative_get(
      endpoint_url='https://api.newrelic.com/v2/alerts_conditions.json',
      response_key='conditions',
      params={'policy_id': policy_id}
    )

  @exponential_retry_on_5xx
  def put_policy_alert_apm_condition(self, condition_id, condition):
    put_condition = requests.put(
      url='https://api.newrelic.com/v2/alerts_conditions/{}.json'.format(condition_id),
      headers=self.auth_header,
      data=json.dumps({"condition": condition})
    )

    self._log_response_errors(put_condition)
    put_condition.raise_for_status()
    return

  def get_applications(self, application_name=''):
    return self.iterative_get(
      endpoint_url='https://api.newrelic.com/v2/applications.json',
      response_key='applications',
      params={'filter[name]': application_name}
    )

  @exponential_retry_on_5xx
  def create_alert_channel(self, name, channel_type, configuration):
    create_channel = requests.post(
      url='https://api.newrelic.com/v2/alerts_channels.json',
      headers=self.auth_header,
      json={
        'channel': {
          'name': name,
          'type': channel_type,
          'configuration': configuration
        }
      })

    self._log_response_errors(create_channel)
    create_channel.raise_for_status()
    return create_channel.json()['channels'][0]

  def create_opsgenie_alert_channel(self, name, api_key, teams=(), tags=(), recipients=()):
    if any([isinstance(arg, basestring) for arg in [teams, tags, recipients]]):
      raise ValueError("teams:{}, tags:{} and recipients:{} should not be of type string:".format(*map(type, [teams, tags, recipients])))

    og_channel_configuration = {
      'api_key': api_key,
      'teams': ','.join(teams),
      'tags': ','.join(tags),
      'recipients': ','.join(recipients)
    }
    return self.create_alert_channel(name, 'opsgenie', og_channel_configuration)

  def get_notification_channel_by_name(self, name):
    alert_channels = self.get_all_notification_channels()
    for channel in alert_channels:
      if channel['name'] == name:
        return channel
    return None

  def get_all_notification_channels(self, refresh=False):
    if self.all_notification_channels and not refresh:
      return self.all_notification_channels

    self.all_notification_channels = self.iterative_get(
      endpoint_url='https://api.newrelic.com/v2/alerts_channels.json',
      response_key='channels'
    )
    return self.all_notification_channels

  def _log_response_errors(self, response):
    """
    If the request does not have a 2XX status response, output the errors
    Args:
      response: response object
    """
    if not response.ok:
      logger.error("Request {} failed for the following reason {}".format(response.url, response.text))



"""
Copyright 2016-2017 Ellation, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import print_function
import os
import sys

import boto3
from botocore.exceptions import ClientError
import yaml

import ef_utils


class EFSiteConfig(object):
  """
  Loads ef_site_config.yml
  """

  def __init__(self):
    self._ef_site_config = os.path.join(os.getcwd(), "ef_site_config.yml")
    self.ssm_parameter_name = os.environ.get('EF_SSM_SITE_CONFIG_LOCATION', '/efopen/ef_site_config')
    self.ssm_region = os.environ.get('EF_SSM_SITE_CONFIG_REGION', 'us-west-2')

  def load(self):
    """Loads the config"""
    whereami = ef_utils.whereami()
    if whereami == 'ec2':
      try:
        return self.load_from_ssm()
      except ClientError as e:
        pass
    return self.load_from_local_file()

  def load_from_ssm(self):
    ssm = boto3.client('ssm', region_name=self.ssm_region)
    parameter_entry = ssm.get_parameter(Name=self.ssm_parameter_name)
    site_config = parameter_entry['Parameter']['Value']
    return yaml.safe_load(site_config)

  def load_from_local_file(self):
    try:
      with open(self._ef_site_config, 'r') as yml_file:
        return yaml.safe_load(yml_file)
    except (IOError, yaml.parser.ParserError) as error:
      print("Error: {}".format(error), file=sys.stderr)
      sys.exit(1)
